from collections.abc import Callable, Iterator, Sequence
from typing import Annotated, TypeVar, Union, overload

from pydantic import BeforeValidator


def _hex_int_validator(value, info):
    if value is None:
        # If not optional, will allow pydantic to (better) handle the error.
        return value

    # NOTE: Allows this module to load lazier.
    from ape.utils.basemodel import ManagerAccessMixin as access

    convert = access.conversion_manager.convert
    return convert(value, int)


HexInt = Annotated[int, BeforeValidator(_hex_int_validator)]
"""
Validate any hex-str or bytes into an integer.
To be used on pydantic-fields.
"""

_T = TypeVar("_T")  # _LazySequence generic.


class _LazySequence(Sequence[_T]):
    def __init__(self, generator: Union[Iterator[_T], Callable[[], Iterator[_T]]]):
        self._generator = generator
        self.cache: list = []

    @overload
    def __getitem__(self, index: int) -> _T: ...

    @overload
    def __getitem__(self, index: slice) -> Sequence[_T]: ...

    def __getitem__(self, index: Union[int, slice]) -> Union[_T, Sequence[_T]]:
        if isinstance(index, int):
            while len(self.cache) <= index:
                # Catch up the cache.
                if value := next(self.generator, None):
                    self.cache.append(value)

            return self.cache[index]

        elif isinstance(index, slice):
            # TODO: Make slices lazier. Right now, it deqeues all.
            for item in self.generator:
                self.cache.append(item)

            return self.cache[index]

        else:
            raise TypeError("Index must be int or slice.")

    def __len__(self) -> int:
        # NOTE: This will deque everything.

        for value in self.generator:
            self.cache.append(value)

        return len(self.cache)

    def __iter__(self) -> Iterator[_T]:
        yield from self.cache
        for value in self.generator:
            yield value
            self.cache.append(value)

    @property
    def generator(self) -> Iterator:
        if callable(self._generator):
            self._generator = self._generator()

        assert isinstance(self._generator, Iterator)  # For type-checking.
        yield from self._generator



from typing import Literal, Union

from pydantic import BaseModel, field_validator


class AutoGasLimit(BaseModel):
    """
    Additional settings for ``gas_limit: auto``.
    """

    multiplier: float = 1.0
    """
    A multiplier to estimated gas.
    """

    @field_validator("multiplier", mode="before")
    @classmethod
    def validate_multiplier(cls, value):
        if isinstance(value, str):
            return float(value)

        return value


GasLimit = Union[Literal["auto", "max"], int, str, AutoGasLimit]
"""
A value you can give to Ape for handling gas-limit calculations.
``"auto"`` refers to automatically figuring out the gas,
``"max"`` refers to using the maximum block gas limit,
and otherwise you can provide a numeric value.
"""



import difflib
import re
import sys
from collections.abc import Iterable
from functools import cached_property
from gettext import gettext
from importlib.metadata import entry_points
from pathlib import Path
from typing import Any, Optional
from warnings import catch_warnings, simplefilter

import click
import yaml

from ape.cli.options import ape_cli_context
from ape.exceptions import Abort, ApeAttributeError, ApeException, handle_ape_exception
from ape.logging import logger

_DIFFLIB_CUT_OFF = 0.6


def display_config(ctx, param, value):
    # NOTE: This is necessary not to interrupt how version or help is intercepted
    if not value or ctx.resilient_parsing:
        return

    from ape.utils.basemodel import ManagerAccessMixin as access

    click.echo("# Current configuration")

    # NOTE: Using json-mode as yaml.safe_dump requires JSON-like structure.
    model = access.local_project.config.model_dump(mode="json")

    click.echo(yaml.safe_dump(model))
    ctx.exit()  # NOTE: Must exit to bypass running ApeCLI


class ApeCLI(click.MultiCommand):
    _CLI_GROUP_NAME = "ape_cli_subcommands"

    def format_commands(self, ctx, formatter) -> None:
        from ape.plugins._utils import PluginMetadataList

        commands = []
        for subcommand in self.list_commands(ctx):
            cmd = self.get_command(ctx, subcommand)
            if cmd is None or cmd.hidden:
                continue

            commands.append((subcommand, cmd))

        if not commands:
            return None

        limit = formatter.width - 6 - max(len(cmd[0]) for cmd in commands)

        # Split the commands into 3 sections.
        sections: dict[str, list[tuple[str, str]]] = {
            "Core": [],
            "Plugin": [],
            "3rd-Party Plugin": [],
        }
        pl_metadata = PluginMetadataList.from_package_names(f"ape_{c[0]}" for c in commands)
        for cli_name, cmd in commands:
            help = cmd.get_short_help_str(limit)
            plugin = pl_metadata.get_plugin(cli_name, check_available=False)
            if plugin is None:
                continue

            if plugin.in_core:
                sections["Core"].append((cli_name, help))
            elif plugin.check_trusted(use_web=False):
                sections["Plugin"].append((cli_name, help))
            else:
                sections["3rd-Party Plugin"].append((cli_name, help))

        for title, rows in sections.items():
            if not rows:
                continue

            with formatter.section(gettext(f"{title} Commands")):
                formatter.write_dl(rows)

    def invoke(self, ctx) -> Any:
        try:
            return super().invoke(ctx)

        except click.UsageError as err:
            self._suggest_cmd(err)

        except ApeException as err:
            path = ctx.obj.local_project.path

            # Extract more interesting ApeException.
            err_to_show = err
            while isinstance(err_to_show, ApeException):
                if (
                    not isinstance(err_to_show, ApeAttributeError)
                    or err_to_show.base_err is None
                    or not isinstance(err_to_show.base_err, ApeException)
                ):
                    break

                err_to_show = err_to_show.base_err

            # NOTE: isinstance check for type-checkers.
            if isinstance(path, Path) and handle_ape_exception(err, (path,)):
                # All exc details already outputted.
                sys.exit(1)
            else:
                raise Abort.from_ape_exception(err) from err

    @staticmethod
    def _suggest_cmd(usage_error):
        if usage_error.message is None:
            raise usage_error

        elif not (match := re.match("No such command '(.*)'.", usage_error.message)):
            raise usage_error

        groups = match.groups()
        if len(groups) < 1:
            raise usage_error

        bad_arg = groups[0]
        suggested_commands = difflib.get_close_matches(
            bad_arg, list(usage_error.ctx.command.commands.keys()), cutoff=_DIFFLIB_CUT_OFF
        )
        if suggested_commands:
            if bad_arg not in suggested_commands:
                usage_error.message = (
                    f"No such command '{bad_arg}'. Did you mean {' or '.join(suggested_commands)}?"
                )

        raise usage_error

    @cached_property
    def commands(self) -> dict:
        _entry_points = entry_points()
        eps: Iterable

        try:
            eps = _entry_points.select(group=self._CLI_GROUP_NAME)
        except AttributeError:
            # Fallback for Python 3.9
            with catch_warnings():
                simplefilter("ignore")
                eps = _entry_points.get(self._CLI_GROUP_NAME, [])  # type: ignore

        commands = {cmd.name.replace("_", "-").replace("ape-", ""): cmd.load for cmd in eps}
        return dict(sorted(commands.items()))

    def list_commands(self, ctx) -> list[str]:
        return [k for k in self.commands]

    def get_command(self, ctx, name) -> Optional[click.Command]:
        try:
            return self.commands[name]()
        except Exception as err:
            logger.warn_from_exception(err, f"Unable to load CLI endpoint for plugin 'ape_{name}'")

        # NOTE: don't return anything so Click displays proper error
        return None


@click.command(cls=ApeCLI, context_settings=dict(help_option_names=["-h", "--help"]))
@ape_cli_context()
@click.version_option(message="%(version)s", package_name="eth-ape")
@click.option(
    "--config",
    is_flag=True,
    is_eager=True,
    expose_value=False,
    callback=display_config,
    help="Show configuration options (using `ape-config.yaml`)",
)
def cli(context):
    _ = context



import copy
import re
from typing import Any, ClassVar, cast

import pytest
from eth_pydantic_types import HexBytes, HexBytes32
from eth_typing import HexAddress, HexStr
from ethpm_types import ContractType, ErrorABI
from ethpm_types.abi import ABIType, EventABI, MethodABI
from evm_trace import CallTreeNode, CallType

from ape.api.networks import ForkedNetworkAPI, NetworkAPI
from ape.exceptions import CustomError, DecodingError, NetworkError, NetworkNotFoundError
from ape.types.address import AddressType
from ape.types.units import CurrencyValueComparable
from ape.utils.misc import DEFAULT_LOCAL_TRANSACTION_ACCEPTANCE_TIMEOUT, LOCAL_NETWORK_NAME
from ape_ethereum.ecosystem import BLUEPRINT_HEADER, BaseEthereumConfig, Block, Ethereum
from ape_ethereum.trace import TransactionTrace
from ape_ethereum.transactions import (
    DynamicFeeTransaction,
    Receipt,
    SetCodeTransaction,
    SharedBlobReceipt,
    SharedBlobTransaction,
    StaticFeeTransaction,
    TransactionType,
)

LOG = {
    "removed": False,
    "logIndex": "0x0",
    "transactionIndex": "0x0",
    "transactionHash": "0x74dd040dfa06f0af9af8ca95d7aae409978400151c746f55ecce19e7356cfc5a",
    "blockHash": "0x2c99950b07accf3e442512a3352a11e6fed37b2331de5f71b7743b357d96e4e8",
    "blockNumber": "0xa946ac",
    "address": "0x274b028b03a250ca03644e6c578d81f019ee1323",
    "data": "0xabffd4675206dab5d04a6b0d62c975049665d1f512f29f303908abdd20bc08a100000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000060000000000000000000000000000000000000000000000000000000000000000744796e616d696300000000000000000000000000000000000000000000000000",
    "topics": [
        "0xa84473122c11e32cd505595f246a28418b8ecd6cf819f4e3915363fad1b8f968",
        "0x0000000000000000000000000000000000000000000000000000000000000006",
        "0x9f3d45ac20ccf04b45028b8080bb191eab93e29f7898ed43acf480dd80bba94d",
    ],
}
CUSTOM_ECOSYSTEM_NAME = "custom-ecosystem"


def make_method_abi(name: str, inputs=None, outputs=None) -> MethodABI:
    inputs = inputs or []
    outputs = outputs or []
    return MethodABI.model_validate(
        {
            "type": "function",
            "name": name,
            "stateMutability": "nonpayable",
            "inputs": inputs,
            "outputs": outputs,
        }
    )


@pytest.fixture
def event_abi(vyper_contract_instance):
    return vyper_contract_instance.NumberChange.abi


@pytest.fixture
def configured_custom_ecosystem(
    custom_networks_config_dict, project, networks, custom_network_name_0
):
    data = copy.deepcopy(custom_networks_config_dict)
    data["networks"]["custom"][0]["ecosystem"] = CUSTOM_ECOSYSTEM_NAME

    # Also ensure we can configure the custom ecosystem as if
    # it were from a plugin.
    data[CUSTOM_ECOSYSTEM_NAME] = {"default_network": custom_network_name_0}

    with project.temp_config(**data):
        yield


def test_name(ethereum):
    assert ethereum.name == "ethereum"


def test_request_header(ethereum):
    actual = ethereum.request_header
    expected = {"User-Agent": "ape-ethereum"}
    assert actual == expected


def test_request_header_subclass():
    class L2(Ethereum):
        name: str = "l2"

    l2 = L2()
    actual = l2.request_header
    expected = {"User-Agent": "ape-l2"}
    assert actual == expected


def test_name_when_custom(configured_custom_ecosystem, networks):
    ecosystem = networks.get_ecosystem(CUSTOM_ECOSYSTEM_NAME)
    actual = ecosystem.name
    expected = CUSTOM_ECOSYSTEM_NAME
    assert actual == expected


@pytest.mark.parametrize(
    "address",
    (
        "0x63953eB1B3D8DB28334E7C1C69456C851F934199".lower(),
        0x63953EB1B3D8DB28334E7C1C69456C851F934199,
    ),
)
def test_decode_address(ethereum, address):
    expected = "0x63953eB1B3D8DB28334E7C1C69456C851F934199"
    actual = ethereum.decode_address(address)
    assert actual == expected


def test_encode_address(ethereum):
    raw_address = "0x63953eB1B3D8DB28334E7C1C69456C851F934199"
    address = AddressType(HexAddress(HexStr(raw_address)))
    actual = ethereum.encode_address(address)
    assert actual == raw_address


def test_encode_calldata(ethereum, address):
    abi = make_method_abi(
        "callMe",
        inputs=[
            ABIType(name="a", type="bytes4"),
            ABIType(name="b", type="address"),
            ABIType(name="c", type="uint256"),
            ABIType(name="d", type="bytes4[]"),
        ],
    )
    byte_array = ["0x456", "0x678"]
    values = ("0x0123", address, HexBytes(55), byte_array)

    actual = ethereum.encode_calldata(abi, *values)
    expected = HexBytes(
        # 0x123
        "0123000000000000000000000000000000000000000000000000000000000000"
        # address
        "000000000000000000000000d8da6bf26964af9d7eed9e03e53415d37aa96045"
        # HexBytes(55)
        "0000000000000000000000000000000000000000000000000000000000000037"
        # byte_array
        "0000000000000000000000000000000000000000000000000000000000000080"
        # length
        "0000000000000000000000000000000000000000000000000000000000000002"
        # first bytes4
        "0456000000000000000000000000000000000000000000000000000000000000"
        # second bytes4
        "0678000000000000000000000000000000000000000000000000000000000000"
    )
    assert actual == expected


@pytest.mark.parametrize(
    "sequence_type,item_type",
    [
        (list, str),
        (tuple, str),
        (list, HexBytes),
        (tuple, HexBytes),
    ],
)
def test_encode_calldata_byte_array(ethereum, sequence_type, item_type):
    """
    Tests against a bug where we could not pass a tuple of HexStr
    for a byte-array.
    """
    abi = make_method_abi(
        "mint", inputs=[{"name": "leaf", "type": "bytes32"}, {"name": "proof", "type": "bytes32[]"}]
    )
    hexstr_array = sequence_type(
        (
            item_type("0xfadbd3"),
            item_type("0x9c6b2c"),
            item_type("0xc5d246"),
        )
    )
    actual = ethereum.encode_calldata(abi, "0x0123", hexstr_array)
    assert isinstance(actual, bytes)


def test_encode_calldata_nested_structs(ethereum):
    abi = make_method_abi(
        "check",
        inputs=[
            ABIType(
                name="data",
                type="tuple",
                components=[
                    ABIType(
                        name="tokenIn", type="address", components=None, internal_type="address"
                    ),
                    ABIType(
                        name="amountIn", type="uint256", components=None, internal_type="uint256"
                    ),
                    ABIType(
                        name="minAmountOut",
                        type="uint256",
                        components=None,
                        internal_type="uint256",
                    ),
                    ABIType(
                        name="path",
                        type="tuple",
                        components=[
                            ABIType(
                                name="pairBinSteps",
                                type="uint256[]",
                                components=None,
                                internal_type="uint256[]",
                            ),
                            ABIType(
                                name="versions",
                                type="uint8[]",
                                components=None,
                                internal_type="enum SwapWrapper.Version[]",
                            ),
                            ABIType(
                                name="tokenPath",
                                type="address[]",
                                components=None,
                                internal_type="address[]",
                            ),
                        ],
                        internal_type="struct SwapWrapper.Path",
                    ),
                ],
                internal_type="struct SwapWrapper.SwapData",
            )
        ],
        outputs=[ABIType(name="", type="bool", components=None, internal_type="bool")],
    )
    calldata = {
        "tokenIn": "0xB31f66AA3C1e785363F0875A1B74E27b85FD66c7",
        "amountIn": 1000000000000000000,
        "minAmountOut": 10000000,
        "path": {
            "pairBinSteps": [0],
            "versions": [0],
            "tokenPath": [
                "0xB31f66AA3C1e785363F0875A1B74E27b85FD66c7",
                "0xB97EF9Ef8734C71904D8002F8b6Bc66Dd9c48a6E",
            ],
        },
    }
    actual = ethereum.encode_calldata(abi, calldata)
    expected = HexBytes(
        "0x000000000000000000000000000000000000000000000000000000000000002"
        "0000000000000000000000000b31f66aa3c1e785363f0875a1b74e27b85fd66c7"
        "0000000000000000000000000000000000000000000000000de0b6b3a76400000"
        "00000000000000000000000000000000000000000000000000000000098968000"
        "00000000000000000000000000000000000000000000000000000000000080000"
        "00000000000000000000000000000000000000000000000000000000000600000"
        "0000000000000000000000000000000000000000000000000000000000a000000"
        "000000000000000000000000000000000000000000000000000000000e0000000"
        "00000000000000000000000000000000000000000000000000000000010000000"
        "00000000000000000000000000000000000000000000000000000000000000000"
        "00000000000000000000000000000000000000000000000000000001000000000"
        "00000000000000000000000000000000000000000000000000000000000000000"
        "00000000000000000000000000000000000000000000000000000200000000000"
        "0000000000000b31f66aa3c1e785363f0875a1b74e27b85fd66c7000000000000"
        "000000000000b97ef9ef8734c71904d8002f8b6bc66dd9c48a6e"
    )
    assert actual == expected


def test_block_handles_snake_case_parent_hash(eth_tester_provider, sender, receiver):
    # Transaction to change parent hash of next block
    sender.transfer(receiver, "1 gwei")

    # Replace 'parentHash' key with 'parent_hash'
    latest_block = eth_tester_provider.get_block("latest")
    latest_block_dict = eth_tester_provider.get_block("latest").model_dump()
    latest_block_dict["parent_hash"] = latest_block_dict.pop("parentHash")

    redefined_block = Block.model_validate(latest_block_dict)
    assert redefined_block.parent_hash == latest_block.parent_hash


def test_transaction_acceptance_timeout(project, networks):
    assert (
        networks.provider.network.transaction_acceptance_timeout
        == DEFAULT_LOCAL_TRANSACTION_ACCEPTANCE_TIMEOUT
    )
    new_value = DEFAULT_LOCAL_TRANSACTION_ACCEPTANCE_TIMEOUT + 1
    timeout_config = {"ethereum": {"local": {"transaction_acceptance_timeout": new_value}}}
    with project.temp_config(**timeout_config):
        assert networks.provider.network.transaction_acceptance_timeout == new_value


def test_decode_logs(ethereum, vyper_contract_instance):
    abi = vyper_contract_instance.NumberChange.abi
    result = [x for x in ethereum.decode_logs([LOG], abi)]
    assert len(result) == 1
    assert result[0].model_dump() == {
        "event_name": "NumberChange",
        "contract_address": "0x274b028b03A250cA03644E6c578D81f019eE1323",
        "event_arguments": {
            "newNum": 6,
            "dynIndexed": HexBytes(
                "0x9f3d45ac20ccf04b45028b8080bb191eab93e29f7898ed43acf480dd80bba94d"
            ),
            "b": HexBytes("0xabffd4675206dab5d04a6b0d62c975049665d1f512f29f303908abdd20bc08a1"),
            "prevNum": 0,
            "dynData": "Dynamic",
        },
        "transaction_hash": "0x74dd040dfa06f0af9af8ca95d7aae409978400151c746f55ecce19e7356cfc5a",
        "block_number": 11093676,
        "block_hash": "0x2c99950b07accf3e442512a3352a11e6fed37b2331de5f71b7743b357d96e4e8",
        "log_index": 0,
        "transaction_index": 0,
    }


def test_decode_logs_empty_list(ethereum, event_abi):
    actual = [x for x in ethereum.decode_logs([], event_abi)]
    assert actual == []


def test_decode_logs_with_struct_from_interface(ethereum):
    abi = EventABI.model_validate(
        {
            "type": "event",
            "name": "ConditionalOrderCreated",
            "inputs": [
                {"name": "owner", "type": "address", "internalType": "address", "indexed": True},
                {
                    "name": "params",
                    "type": "tuple",
                    "components": [
                        {
                            "name": "handler",
                            "type": "address",
                            "internalType": "contract IConditionalOrder",
                        },
                        {"name": "salt", "type": "bytes32", "internalType": "bytes32"},
                        {"name": "staticInput", "type": "bytes", "internalType": "bytes"},
                    ],
                    "internalType": "struct IConditionalOrder.ConditionalOrderParams",
                    "indexed": False,
                },
            ],
            "anonymous": False,
        }
    )
    logs = [
        {
            "address": "0xfdafc9d1902f4e0b84f65f49f244b32b31013b74",
            "blockHash": "0xec1b8e18412dd91114c0b3ea03c41d02243d60cd05661861df5765fb5ec462c6",
            "blockNumber": "0x11101cc",
            "data": "0x00000000000000000000000000000000000000000000000000000000000000200000000000000000000000006cf1e9ca41f7611def408122793c358a3d11e5a5000000000000000000000000000000000000000000000000000000189e491e96000000000000000000000000000000000000000000000000000000000000006000000000000000000000000000000000000000000000000000000000000001400000000000000000000000006b175474e89094c44da98b954eedeac495271d0f000000000000000000000000c02aaa39b223fe8d0a0e5c4f27ead9083c756cc2000000000000000000000000e7602ca44f83a5e9ba8bd14125ddcb295f3d63bd00000000000000000000000000000000000000000000018eb03406b6af33c8c3000000000000000000000000000000000000000000000000345d7669a1e8815700000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000002000000000000000000000000000000000000000000000000000000000000012c00000000000000000000000000000000000000000000000000000000000000007b9bc44550e2cc369010139a1d3a5a6ec6d65910379e0ee35a7c3cdf24cb12a7",
            "logIndex": "0x97",
            "removed": False,
            "topics": [
                "0x2cceac5555b0ca45a3744ced542f54b56ad2eb45e521962372eef212a2cbf361",
                "0x000000000000000000000000e7602ca44f83a5e9ba8bd14125ddcb295f3d63bd",
            ],
            "transactionHash": "0x55859e74e591100e0f37f146abc0dbebe2252eaf5f533319ca77427745b85ceb",
            "transactionIndex": "0x45",
        }
    ]
    actual = list(ethereum.decode_logs(logs, abi))
    assert len(actual) == 1
    assert actual[0].owner == "0xE7602Ca44f83a5E9Ba8BD14125dDcb295f3D63BD"
    assert actual[0].params == [
        "0x6cf1e9ca41f7611def408122793c358a3d11e5a5",
        HexBytes("0x000000000000000000000000000000000000000000000000000000189e491e96"),
        HexBytes(
            "0x0000000000000000000000006b175474e89094c44da98b954eedeac495271d0f0000000"
            "00000000000000000c02aaa39b223fe8d0a0e5c4f27ead9083c756cc20000000000000000"
            "00000000e7602ca44f83a5e9ba8bd14125ddcb295f3d63bd0000000000000000000000000"
            "0000000000000000000018eb03406b6af33c8c30000000000000000000000000000000000"
            "00000000000000345d7669a1e881570000000000000000000000000000000000000000000"
            "0000000000000000000000000000000000000000000000000000000000000000000000000"
            "0000000000020000000000000000000000000000000000000000000000000000000000000"
            "12c00000000000000000000000000000000000000000000000000000000000000007b9bc4"
            "4550e2cc369010139a1d3a5a6ec6d65910379e0ee35a7c3cdf24cb12a7"
        ),
    ]


def test_decode_block_when_hash_is_none(ethereum):
    # When using some providers, such as hardhat, the hash of the pending block is None
    block_data_with_none_hash: dict[str, Any] = {
        "number": None,
        "hash": None,
        "parentHash": HexBytes(
            "0xcb94e150c06faee9ab2bf12a40b0937ac9eab1879c733ebe3249aafbba2f80b1"
        ),
        "nonce": None,
        "mixHash": None,
        "sha3Uncles": HexBytes(
            "0x1dcc4de8dec75d7aab85b567b6ccd41ad312451b948a7413f0a142fd40d49347"
        ),
        "logsBloom": None,
        "transactionsRoot": HexBytes(
            "0x56e81f171bcc55a6ff8345e692c0f86e5b48e01b996cadc001622fb5e363b421"
        ),
        "stateRoot": HexBytes("0x8728474146565003152f9cee496de043fd68566dabdb06116a0d5bfc63e1a5a9"),
        "receiptsRoot": HexBytes(
            "0x56e81f171bcc55a6ff8345e692c0f86e5b48e01b996cadc001622fb5e363b421"
        ),
        "miner": "0xC014BA5EC014ba5ec014Ba5EC014ba5Ec014bA5E",
        "difficulty": 131072,
        "totalDifficulty": 131073,
        "extraData": HexBytes("0x"),
        "size": 513,
        "gasLimit": 0,
        "gasUsed": 0,
        "timestamp": 1660932629,
        "transactions": [],
        "uncles": [],
        "baseFeePerGas": 0,
    }
    actual = ethereum.decode_block(block_data_with_none_hash)
    assert actual.hash is None
    assert actual.gas_limit == 0


def test_decode_block_with_hex_values(ethereum):
    """
    When using WS providers directly, the values are all hex strings (more raw).
    This test ensures we can still decode blocks that are in this form.
    """
    raw_block = {
        "baseFeePerGas": "0x62d76fae2",
        "difficulty": "0x0",
        "extraData": "0x506f776572656420627920626c6f58726f757465",
        "gasLimit": "0x1c9c380",
        "gasUsed": "0x91b5e5",
        "hash": "0x925cac44d8bac5df3e3eeed6379b3924d6769a054b3c4079899c1d9b442a4041",
        "logsBloom": "0x7823d304ab8c72f91270a023c11482f341a08a00c91b00f8859906c8e9f22e04b19d311e422cfa79d8107e2b901fc7d9c62fcd2d8a673802f500f4a023693525a2ce66481cc51d1a4e97422b8424912b8c719c6821589a240d5015c5cbe4de46b3b60ad43621cd2f540119b3c938d88f011a49d755a31cc415f34896596912d40b0397d0c544571704ea001c56303f02086e16a3a3b7168f443d224034102ca4c6b21456932dae401c8da7c51d36ae3804c8a084811802c708ae94b7010302100700a28f156d2ca2d61420320e82e8b9927b626bc0ec003dfa09009e3d1668798fb0f8c81a205eb100ee8ea2e20c6041b2d8b2b01bd4d14e8acfaa317049d654",
        "miner": "0x6b3a8798e5fb9fc5603f3ab5ea2e8136694e55d0",
        "mixHash": "0x0493af8ac5b7de66c7660c22a703a54069cbd561a2767537eaef994bd77e084b",
        "nonce": "0x0000000000000000",
        "number": "0x11234df",
        "parentHash": "0x47651c13bda353d1bbc610bbdb8a9aa1436629f05f8a32a27f143200da698df8",
        "receiptsRoot": "0xa5abe7eeac74aa03ea82421fcef9b74d1822cfe0ee4b29a8dc01eef31e7d0019",
        "sha3Uncles": "0x1dcc4de8dec75d7aab85b567b6ccd41ad312451b948a7413f0a142fd40d49347",
        "size": "0x1cf02",
        "stateRoot": "0x59eb1dfb098361d1f74dffc8436da680715f671dc3f02be28f4db3f44b1a09bf",
        "timestamp": "0x64e4b0cb",
        "transactionsRoot": "0xf2e3c3e4fbe06ff882f3deb8a7ec8cbbe09e55b2c995d93dd0a36d5e843f3efc",
    }
    actual = ethereum.decode_block(raw_block)
    assert actual.difficulty == 0


def test_decode_logs_topics_not_first(ethereum):
    """
    Tests against a condition where we could not decode logs if the
    topics were not defined first.
    """
    abi_dict = {
        "anonymous": False,
        "inputs": [
            {"indexed": False, "internalType": "address", "name": "sender", "type": "address"},
            {"indexed": True, "internalType": "address", "name": "owner", "type": "address"},
            {"indexed": True, "internalType": "int24", "name": "tickLower", "type": "int24"},
            {"indexed": True, "internalType": "int24", "name": "tickUpper", "type": "int24"},
            {"indexed": False, "internalType": "uint128", "name": "amount", "type": "uint128"},
            {"indexed": False, "internalType": "uint256", "name": "amount0", "type": "uint256"},
            {"indexed": False, "internalType": "uint256", "name": "amount1", "type": "uint256"},
        ],
        "name": "Mint",
        "type": "event",
    }
    abi = EventABI.model_validate(abi_dict)
    logs = [
        {
            "address": "0x3416cf6c708da44db2624d63ea0aaef7113527c6",
            "blockHash": "0x488f23ba55f64bf1aac02ee7278b70c3f4bb2fb57b8aaa6ab3b481f1809f18ea",
            "blockNumber": "0xcfa869",
            "data": "0x000000000000000000000000c36442b4a4522e871399cd717abdd847ab11fe8800000000000000000000000000000000000000000000000000821d6b102660fb000000000000000000000000000000000000000000000000000009fde8545338000000000000000000000000000000000000000000000000000003548cf35031",
            "logIndex": "0x7a",
            "removed": False,
            "topics": [
                "0x7a53080ba414158be7ec69b987b5fb7d07dee101fe85488f0853ae16239d0bde",
                "0x000000000000000000000000c36442b4a4522e871399cd717abdd847ab11fe88",
                "0xfffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffc",
                "0x0000000000000000000000000000000000000000000000000000000000000004",
            ],
            "transactionHash": "0xf093a630478562d03e3b2476a5b0551609722747d442a462579fa02e1332a941",
            "transactionIndex": "0x41",
        }
    ]
    actual = list(ethereum.decode_logs(logs, abi))
    assert actual


def test_decode_receipt(eth_tester_provider, ethereum):
    receipt_data = {
        "provider": eth_tester_provider,
        "required_confirmations": 0,
        "blockHash": HexBytes("0x8988adc8a0b346526f1d769841b7464e38a282b25ed346f1f810be8cb7393bc2"),
        "blockNumber": 2,
        "from": "0x318b469BBa396AEc2C60342F9441be36A1945174",
        "gas": 1764431,
        "hash": HexBytes("0x58d086e6b7ecf55c92a5fe420d870ff2eebe99824a711a3cdbe29f497a0e534c"),
        "input": " ",
        "nonce": 1,
        "to": None,
        "transactionIndex": 0,
        "value": 0,
        "v": 1,
        "r": HexBytes("0xfa532141efbfa5bbeaa542cf68bfe5df97dd18b696b8a183f74714758a646238"),
        "s": HexBytes("0x1b80edf76bb6f0295ff847c1b99eb2928ea3533fe41125cfe0fa17fdc634b938"),
        "type": 2,
        "accessList": [],
        "chainId": 31337,
        "gasPrice": 0,
        "maxFeePerGas": 0,
        "maxPriorityFeePerGas": 0,
        "transactionHash": HexBytes(
            "0x58d086e6b7ecf55c92a5fe420d870ff2eebe99824a711a3cdbe29f497a0e534c"
        ),
        "cumulativeGasUsed": 1764431,
        "gasUsed": 1764431,
        "contractAddress": "0xC072c85922b7233998B2Ab990fEFdE80218Ca63F",
        "logs": [],
        "logsBloom": HexBytes(
            "0x00000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000"
        ),
        "status": 1,
        "effectiveGasPrice": 0,
    }
    receipt = ethereum.decode_receipt(receipt_data)
    assert type(receipt) is Receipt  # NOTE: Purposely not using isinstance()

    # Tests against bug where input data would come back improperly
    assert receipt.data == HexBytes(
        " "
    )


def test_decode_receipt_from_etherscan(eth_tester_provider, ethereum):
    receipt = ethereum.decode_receipt(
        {
            "blockNumber": "11291970",
            "timeStamp": "1661846925",
            "hash": "0x5780b43d819035ed1fa079171bdce7f0bbeaa6b01f201f8985d279a66cfc6844",
            "nonce": "0",
            "blockHash": "0x3175f953c1da4bf3d15b853dae4a150ae44e2e71380936463e89142c12961968",
            "transactionIndex": "31",
            "from": "0xf39Fd6e51aad88F6F4ce6aB8827279cffFb92266",
            "to": "0xe7f1725E7734CE288F8367e1Bb143E90bb3F0512",
            "value": "0",
            "gas": "4712388",
            "gasPrice": "1499999989",
            "isError": "0",
            "input": "0x608060405234801561000f575f80fd5b506101248061001d5f395ff3fe6080604052348015600e575f80fd5b50600436106026575f3560e01c8063f207564e14602a575b5f80fd5b6039603536600460b4565b603b565b005b6040516390b5561d60e01b81526004810182905273274b028b03a250ca03644e6c578d81f019ee1323906390b5561d90602401602060405180830381865af41580156088573d5f803e3d5ffd5b505050506040513d601f19601f8201168201806040525081019060aa919060ca565b60b1575f80fd5b50565b5f6020828403121560c3575f80fd5b5035919050565b5f6020828403121560d9575f80fd5b8151801515811460e7575f80fd5b939250505056fea2646970667358221220050eefcb4ae3417fbe131ccad8d577d9488919884d4c4a4d0bf8b9c46ce640f664736f6c63430008150033",
            "contractAddress": "",
            "cumulativeGasUsed": "8978461",
            "gasUsed": "257131",
            "methodId": "0xf926657e",
            "functionName": "setup(address _reputationOracle, address _recordingOracle, uint256 _reputationOracleStake, uint256 _recordingOracleStake, string _url, string _hash)",
            "required_confirmations": "13703",
            "status": "1",
            "chainId": 1,
        }
    )
    assert type(receipt) is Receipt  # NOTE: Purposely not using isinstance()
    assert receipt.type == 0
    assert receipt.max_fee > 0
    assert receipt.gas_price == 1499999989


@pytest.mark.parametrize(
    "blob_gas_used,blob_gas_key",
    [
        ("0x20000", "blobGasUsed"),
        (131072, "blob_gas_used"),
        (0, "blobGasUsed"),
        (None, "blobGasUsed"),
    ],
)
def test_decode_receipt_shared_blob(ethereum, blob_gas_used, blob_gas_key):
    blob_gas_price = "0x4d137e31b"

    data = {
        "required_confirmations": 0,
        "blockHash": HexBytes("0x051fb508617fcbe0034538b7ee54c1dedbbbaa6f8d0aeb776edf81bafbc883bd"),
        "blockNumber": 10526428,
        "from": "0x194E22F49BC3f58903866d55488E1e9e8d69b517",
        "gas": 5500000,
        "gasPrice": 1000000008,
        "maxFeePerGas": 150000000000,
        "maxPriorityFeePerGas": 1000000000,
        "maxFeePerBlobGas": "0x22ecb25c00",
        "hash": HexBytes("0xd2882bae0d79a6c8e0fbf0089bbcb4b2eef3a1365471ad9f779b06a41ba47d3c"),
        "input": HexBytes("0xb72d42a1000000000000000000000000000000000000000000"),  # Note: abridged
        "nonce": 329925,
        "to": "0xd5c325D183C592C94998000C5e0EED9e6655c020",
        "transactionIndex": 48,
        "value": 0,
        "type": 3,
        "accessList": [],
        "chainId": 5,
        "blobVersionedHashes": [
            "0x015405d502a27897ad38ffcb80566d1559fa53764befc6d90731082837c2d19e"
        ],
        "v": 0,
        "r": HexBytes("0x4e5bdcbba31548cb8f9806491c5ced7b0f1eac78c7dc0677616c23ee0e469f74"),
        "s": HexBytes("0x31c98ea044c97225d83b9a3f0fa719e2299b9fbc6436e4b3eb096ea528de03ff"),
        "yParity": 0,
        "blobGasPrice": blob_gas_price,
        blob_gas_key: blob_gas_used,
        "contractAddress": None,
        "cumulativeGasUsed": 23085827,
        "effectiveGasPrice": 1000000008,
        "gasUsed": 219756,
        "logs": [],
        "logsBloom": HexBytes("0x00000000010002"),
        "status": 1,
        "transactionHash": HexBytes(
            "0xd2882bae0d79a6c8e0fbf0089bbcb4b2eef3a1365471ad9f779b06a41ba47d3c"
        ),
    }
    actual = ethereum.decode_receipt(data)
    assert isinstance(actual, SharedBlobReceipt)
    assert actual.blob_gas_price == int(blob_gas_price, 16)

    if blob_gas_used:
        # all test-values are this when they exist.
        assert actual.blob_gas_used == 131072
    else:
        # when None, should also default to 0.
        assert actual.blob_gas_used == 0

    # Show type=3 is required.
    data["type"] = 2
    actual = ethereum.decode_receipt(data)
    assert not isinstance(actual, SharedBlobReceipt)
    assert isinstance(actual, Receipt)


def test_decode_receipt_misleading_blob_receipt(ethereum):
    """
    Tests a strange situation (noticed on Tenderly nodes) where _some_
    of the keys indicate blob-related fields, set to ``0``, and others
    are missing, because it's not actually a blob receipt. In this case,
    don't use the blob-receipt class.
    """
    data = {
        "type": 2,
        "status": 1,
        "cumulativeGasUsed": 10565720,
        "logsBloom": HexBytes(
            "0x00000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000"
        ),
        "logs": [],
        "transactionHash": HexBytes(
            "0x62fc9991bc7fb0c76bc83faaa8d1c17fc5efb050542e58ac358932f80aa7a087"
        ),
        "from": "0x1f9090aaE28b8a3dCeaDf281B0F12828e676c326",
        "to": "0xeBec795c9c8bBD61FFc14A6662944748F299cAcf",
        "contractAddress": None,
        "gasUsed": 21055,
        "effectiveGasPrice": 7267406643,
        "blockHash": HexBytes("0xa47fc133f829183b751488c1146f1085451bcccd247db42066dc6c89eaf5ebac"),
        "blockNumber": 21051245,
        "transactionIndex": 130,
        "blobGasUsed": 0,
    }
    actual = ethereum.decode_receipt(data)
    assert not isinstance(actual, SharedBlobReceipt)
    assert isinstance(actual, Receipt)


def test_default_transaction_type_not_connected_used_default_network(project, ethereum, networks):
    value = TransactionType.STATIC.value
    config_dict = {"ethereum": {"mainnet_fork": {"default_transaction_type": value}}}
    assert ethereum.default_transaction_type == TransactionType.DYNAMIC

    with project.temp_config(**config_dict):
        ethereum._default_network = "mainnet-fork"
        provider = networks.active_provider

        # Disconnect so it uses default.
        networks.active_provider = None
        try:
            assert ethereum.default_network_name == "mainnet-fork"
            assert ethereum.default_transaction_type == TransactionType.STATIC
            ethereum._default_network = LOCAL_NETWORK_NAME
        finally:
            networks.active_provider = provider


def test_default_transaction_type_configured_from_local_network(
    eth_tester_provider, ethereum, project
):
    _ = eth_tester_provider  # Connection required so 'ethereum' knows the network.
    value = TransactionType.STATIC.value
    config = {"ethereum": {LOCAL_NETWORK_NAME: {"default_transaction_type": value}}}
    with project.temp_config(**config):
        assert ethereum.default_transaction_type == TransactionType.STATIC


def test_default_transaction_type_changed_at_class_level(ethereum):
    """
    Simulates an L2 plugin changing the default at the definition-level.
    """

    class L2NetworkConfig(BaseEthereumConfig):
        DEFAULT_TRANSACTION_TYPE: ClassVar[int] = TransactionType.STATIC.value

    config = L2NetworkConfig()
    assert config.local.default_transaction_type.value == 0
    assert config.mainnet.default_transaction_type.value == 0
    assert config.mainnet_fork.default_transaction_type.value == 0


@pytest.mark.parametrize("network_name", (LOCAL_NETWORK_NAME, "mainnet-fork", "mainnet_fork"))
def test_gas_limit_local_networks(ethereum, network_name):
    network = ethereum.get_network(network_name)
    network.__dict__.pop("gas_limit", None)  # Refresh in case was changed in another test.
    assert network.gas_limit == "max"


def test_gas_limit_live_networks(ethereum):
    network = ethereum.get_network("sepolia")
    assert network.gas_limit == "auto"


def test_encode_blueprint_contract(ethereum, vyper_contract_type):
    actual = ethereum.encode_contract_blueprint(vyper_contract_type)
    ct_bytes = vyper_contract_type.deployment_bytecode.to_bytes()

    # EIP-5202
    expected = (
        HexBytes("0x61")
        + (len(ct_bytes) + 3).to_bytes(2, "big")  # preface and length
        + HexBytes("0x3d81600a3d39f3")  # return stuff
        + BLUEPRINT_HEADER
        + HexBytes(0)
        + ct_bytes
    )

    assert actual.data == HexBytes(expected)


def test_decode_returndata(ethereum):
    abi = make_method_abi("doThing", outputs=[{"name": "", "type": "bool"}])
    data = HexBytes32.__eth_pydantic_validate__(0)
    actual = ethereum.decode_returndata(abi, data)
    assert actual == (False,)


def test_decode_returndata_non_empty_padding_bytes(ethereum):
    raw_data = HexBytes(
        "0x08c379a000000000000000000000000000000000000000000000000000000000000000200"
        "000000000000000000000000000000000000000000000000000000000000012696e73756666"
        "696369656e742066756e64730000000000000000000000000000"
    )
    abi = make_method_abi(
        "transfer",
        inputs=[
            {"name": "receiver", "type": "address"},
            {"name": "amount", "type": "uint256"},
        ],
        outputs=[{"name": "", "type": "bool"}],
    )
    with pytest.raises(DecodingError):
        ethereum.decode_returndata(abi, raw_data)


def test_decode_returndata_no_bytes_returns_zero(ethereum):
    abi = make_method_abi("doThing", outputs=[{"name": "", "type": "bool"}])
    actual = ethereum.decode_returndata(abi, b"")
    assert actual == (0,)


def test_decode_returndata_list_with_1_struct(ethereum):
    """
    Tests a condition where an array of a list with 1 struct
    would be turned into a raw tuple instead of the Struct class.
    """
    abi = make_method_abi(
        "getArrayOfStructs",
        outputs=[
            ABIType(
                name="",
                type="tuple[]",
                components=[
                    ABIType(name="a", type="address", components=None, internal_type=None),
                    ABIType(name="b", type="bytes32", components=None, internal_type=None),
                ],
                internal_type=None,
            )
        ],
    )
    raw_data = HexBytes(
        "0x000000000000000000000000000000000000000000000000000000000000002"
        "00000000000000000000000000000000000000000000000000000000000000001"
        "000000000000000000000000f39fd6e51aad88f6f4ce6ab8827279cfffb92266f"
        "d91dc47b758f65fd0f6fe511566770c0ae3f94ff9999ceb23bfec3ac9fdc168"
    )
    actual = ethereum.decode_returndata(abi, raw_data)

    assert len(actual) == 1
    # The result should still be a list.
    # There was also a bug where it was mistakenly a tuple!
    assert isinstance(actual[0], list)
    assert len(actual[0]) == 1
    assert actual[0][0].a == "0xf39Fd6e51aad88F6F4ce6aB8827279cffFb92266"
    assert actual[0][0].b == HexBytes(
        "0xfd91dc47b758f65fd0f6fe511566770c0ae3f94ff9999ceb23bfec3ac9fdc168"
    )


def test_decode_returndata_returns_str_comparable_ints(ethereum):
    abi = make_method_abi(
        "gimmeInts",
        outputs=[ABIType(name="", type="int", components=None, internal_type="int")],
    )
    raw_data = HexBytes("0x000000000000000000000000000000000000000000000000000000000000002")
    actual = ethereum.decode_returndata(abi, raw_data)
    assert isinstance(actual[0], int)
    assert isinstance(actual[0], CurrencyValueComparable)


def test_decode_returndata_bool(ethereum):
    abi = MethodABI(
        type="function",
        name="view_method",
        stateMutability="view",
        inputs=[],
        outputs=[ABIType(name="", type="bool", components=None, internal_type=None)],
    )
    raw_data = HexBytes("0x000000000000000000000000000000000000000000000000000000000000001")
    actual = ethereum.decode_returndata(abi, raw_data)[0]
    assert isinstance(actual, bool)


@pytest.mark.parametrize("tx_type", TransactionType)
def test_create_transaction_uses_network_gas_limit(tx_type, ethereum, eth_tester_provider, owner):
    tx = ethereum.create_transaction(type=tx_type.value, sender=owner.address)
    assert tx.type == tx_type.value
    assert tx.gas_limit == eth_tester_provider.max_gas


def test_create_transaction_with_none_values(ethereum, eth_tester_provider):
    """
    Tests against None being in place of values in kwargs,
    causing their actual defaults not to get used and ValidationErrors
    to occur.
    """
    static = ethereum.create_transaction(
        value=None, data=None, chain_id=None, gas_price=None, nonce=None, receiver=None
    )
    dynamic = ethereum.create_transaction(
        value=None,
        data=None,
        chain_id=None,
        max_fee=None,
        max_priority_fee=None,
        nonce=None,
        receiver=None,
    )
    assert isinstance(static, StaticFeeTransaction)  # Because used gas_price
    assert isinstance(dynamic, DynamicFeeTransaction)  # Because used max_fee

    for tx in (static, dynamic):
        assert tx.data == b""  # None is not allowed.
        assert tx.value == 0  # None is same as 0.
        assert tx.chain_id == eth_tester_provider.chain_id
        assert tx.nonce is None

    assert static.gas_price is None
    assert dynamic.max_fee is None
    assert dynamic.max_priority_fee is None


@pytest.mark.parametrize("kwarg_name", ("max_fee", "max_fee_per_gas", "maxFee", "maxFeePerGas"))
def test_create_transaction_max_fee_per_gas(kwarg_name, ethereum):
    fee = 1_000_000_000
    tx = ethereum.create_transaction(**{kwarg_name: fee})
    assert tx.max_fee == fee


@pytest.mark.parametrize("kwarg_name", ("gas_price", "gasPrice"))
def test_create_transaction_with_gas_price(kwarg_name, ethereum):
    price = 123
    tx = ethereum.create_transaction(**{kwarg_name: price})
    assert tx.gas_price == price


@pytest.mark.parametrize(
    "tx_kwargs",
    [{"type": 3}, {"max_fee_per_blob_gas": 123}, {"blob_versioned_hashes": [HexBytes(123)]}],
)
def test_create_transaction_shared_blob(tx_kwargs, ethereum):
    tx = ethereum.create_transaction(**tx_kwargs)
    assert isinstance(tx, SharedBlobTransaction)
    assert tx.type == TransactionType.SHARED_BLOB.value


@pytest.mark.parametrize(
    "kwarg_name,value", [("max_fee_per_blob_gas", 345), ("maxFeePerBlobGas", "0x2343")]
)
def test_create_transaction_max_fee_per_blob_gas(kwarg_name, value, ethereum):
    tx = ethereum.create_transaction(**{kwarg_name: value})
    assert isinstance(tx, SharedBlobTransaction)
    expected = value if isinstance(value, int) else int(value, 16)
    assert tx.max_fee_per_blob_gas == expected


@pytest.mark.parametrize(
    "kwarg_name,value",
    [("blob_versioned_hashes", "0x123"), ("blobVersionedHashes", HexBytes("0x123"))],
)
def test_create_transaction_blob_versioned_hashed(kwarg_name, value, ethereum):
    tx = ethereum.create_transaction(**{kwarg_name: [value]})
    assert isinstance(tx, SharedBlobTransaction)
    assert tx.blob_versioned_hashes == [HexBytes(value)]


@pytest.mark.parametrize(
    "tx_kwargs",
    [{"type": 4}, {"authorizationList": []}],
)
def test_create_transaction_set_code(tx_kwargs, ethereum):
    tx = ethereum.create_transaction(**tx_kwargs)
    assert isinstance(tx, SetCodeTransaction)
    assert tx.type == TransactionType.SET_CODE.value


@pytest.mark.parametrize("tx_type", TransactionType)
def test_encode_transaction(tx_type, ethereum, vyper_contract_instance, owner, eth_tester_provider):
    abi = vyper_contract_instance.contract_type.methods[0]
    actual = ethereum.encode_transaction(
        vyper_contract_instance.address, abi, sender=owner.address, type=tx_type.value
    )
    assert actual.gas_limit == eth_tester_provider.max_gas


def test_set_default_network_not_exists(ethereum):
    bad_network = "NOT_EXISTS"
    expected = f"No network in 'ethereum' named '{bad_network}'. Options:.*"
    with pytest.raises(NetworkNotFoundError, match=expected):
        ethereum.set_default_network(bad_network)


def test_networks(ethereum):
    actual = ethereum.networks

    with_forks = ("sepolia", "mainnet")
    without_forks = (LOCAL_NETWORK_NAME,)

    def assert_net(
        net: str,
    ):
        assert net in actual
        base_class = ForkedNetworkAPI if net.endswith("-fork") else NetworkAPI
        assert isinstance(actual[net], base_class)

    for net in with_forks:
        assert_net(net)
        assert_net(f"{net}-fork")
    for net in without_forks:
        assert_net(net)


def test_networks_includes_custom_networks(
    ethereum, custom_networks_config_dict, project, custom_network_name_0, custom_network_name_1
):
    with project.temp_config(**custom_networks_config_dict):
        actual = ethereum.networks
        for net in (
            "sepolia",
            "mainnet",
            LOCAL_NETWORK_NAME,
            custom_network_name_0,
            custom_network_name_1,
            # Show custom networks are auto-forked!
            f"{custom_network_name_0}-fork",
            f"{custom_network_name_1}-fork",
        ):
            assert net in actual
            assert isinstance(actual[net], NetworkAPI)


def test_networks_when_custom_ecosystem(
    configured_custom_ecosystem, networks, custom_network_name_0
):
    obj = networks.custom_ecosystem
    actual = obj.networks
    assert obj.name == CUSTOM_ECOSYSTEM_NAME
    assert custom_network_name_0 in actual
    assert "mainnet" not in actual


def test_networks_multiple_networks_with_same_name(custom_networks_config_dict, ethereum, project):
    data = copy.deepcopy(custom_networks_config_dict)
    data["networks"]["custom"][0]["name"] = "foonet"
    data["networks"]["custom"][1]["name"] = "foonet"
    expected = ".*More than one network named 'foonet' in ecosystem 'ethereum'.*"
    with project.temp_config(**data):
        with pytest.raises(NetworkError, match=expected):
            _ = ethereum.networks


def test_getattr(ethereum):
    assert ethereum.mainnet.name == "mainnet"
    assert isinstance(ethereum.mainnet, NetworkAPI)


def test_getattr_custom_networks(
    ethereum, custom_networks_config_dict, project, custom_network_name_0
):
    with project.temp_config(**custom_networks_config_dict):
        actual = getattr(ethereum, custom_network_name_0)
        assert actual.name == custom_network_name_0
        assert isinstance(actual, NetworkAPI)


def test_default_network(ethereum):
    assert ethereum.default_network_name == LOCAL_NETWORK_NAME


def test_default_network_when_custom_and_set_in_config(
    configured_custom_ecosystem, networks, custom_network_name_0
):
    ecosystem = networks.get_ecosystem(CUSTOM_ECOSYSTEM_NAME)
    # Force it to use config value (in case was set from previous test)
    ecosystem._default_network = None
    assert ecosystem.default_network_name == custom_network_name_0


def test_default_network_name_set_programmatically(ethereum):
    ethereum._default_network = "testnet"
    assert ethereum.default_network_name == "testnet"
    ethereum._default_network = None


def test_default_network_name_from_config(project, ethereum):
    cfg = {"ethereum": {"default_network": "sepolia"}}
    ethereum._default_network = None
    with project.temp_config(**cfg):
        assert ethereum.default_network_name == "sepolia"


def test_default_network_name_when_not_set_uses_local(project, ethereum):
    orig = project.config.ethereum
    data = orig if isinstance(orig, dict) else orig.model_dump()
    data = {k: v for k, v in data.items() if k not in ("default_network",)}
    data["default_network"] = None
    with project.temp_config(**data):
        ethereum._default_network = None
        assert ethereum.default_network_name == LOCAL_NETWORK_NAME


def test_default_network_name_when_not_set_and_no_local_uses_only(
    project, custom_networks_config_dict
):
    """
    Tests a condition that is rare but when a default network is
    not set but there is a single network. In this case, it
    should use the single network by default.
    """
    net = copy.deepcopy(custom_networks_config_dict["networks"]["custom"][0])

    # In a situation with an ecosystem with only a single network.
    ecosystem_name = "acustomeco"
    net["ecosystem"] = ecosystem_name

    only_network = "onlynet"  # More obvious name for test.
    net["name"] = only_network

    with project.temp_config(networks={"custom": [net]}):
        ecosystem = project.network_manager.get_ecosystem(ecosystem_name)
        ecosystem._default_network = None
        actual = ecosystem.default_network_name

        if actual == LOCAL_NETWORK_NAME:
            # For some reason, this test is flake-y. Offer more info
            # to try and debug when this happens (intermittent CI failure).
            all_nets = ", ".join([x for x in ecosystem.networks.keys()])
            pytest.fail(
                f"assert '{LOCAL_NETWORK_NAME}' == '{only_network}'. More info below:\n"
                f"ecosystem_name={ecosystem.name}\n"
                f"networks={all_nets}\n"
                f"type={type(ecosystem)}"
            )
        else:
            # This should pass.
            assert ecosystem.default_network_name == only_network


def test_decode_custom_error(chain, ethereum):
    data = HexBytes("0x6a12f104")
    abi = [ErrorABI(type="error", name="InsufficientETH")]
    contract_type = ContractType(abi=abi)
    addr = cast(AddressType, "0x3fC91A3afd70395Cd496C647d5a6CC9D4B2b7FAD")

    # Hack in contract-type.
    chain.contracts[addr] = contract_type

    actual = ethereum.decode_custom_error(data, addr)
    assert isinstance(actual, CustomError)


def test_decode_custom_error_contract_type_not_found(ethereum):
    data = HexBytes("0x6a12f104")
    # not known
    addr = cast(AddressType, "0x4fC92A3afd70395Cd496C647d5a6CC9D4B2b7FAD")
    actual = ethereum.decode_custom_error(data, addr)
    assert actual is None


def test_decode_custom_error_tx_unsigned(ethereum):
    data = HexBytes("0x6a12f104")
    # not known
    addr = cast(AddressType, "0x4fC92A3afd70395Cd496C647d5a6CC9D4B2b7FAD")
    actual = ethereum.decode_custom_error(data, addr)
    assert actual is None


def test_decode_custom_error_selector_not_found(chain, ethereum):
    data = HexBytes("0x6a12f104")
    abi: list = []
    contract_type = ContractType(abi=abi)
    addr = cast(AddressType, "0x3fC91A3afd70395Cd496C647d5a6CC9D4B2b7FAD")

    # Hack in contract-type.
    chain.contracts.contract_types[addr] = contract_type

    tx = ethereum.create_transaction()
    actual = ethereum.decode_custom_error(data, addr, txn=tx)
    assert actual is None


def test_enrich_trace(ethereum, vyper_contract_instance, owner):
    tx = vyper_contract_instance.setNumber(96247783, sender=owner)
    trace = TransactionTrace(transaction_hash=tx.txn_hash)
    actual = ethereum.enrich_trace(trace)
    assert isinstance(actual, TransactionTrace)
    assert actual._enriched_calltree is not None
    assert re.match(r"VyperContract\.setNumber\(num=\d*\) \[\d* gas]", repr(actual))


def test_enrich_trace_handles_call_type_enum(ethereum, vyper_contract_instance, owner):
    """
    Testing a custom trace whose call tree uses an Enum type instead of a str.
    """

    class PluginTxTrace(TransactionTrace):
        def get_calltree(self) -> CallTreeNode:
            call = super().get_calltree()
            # Force enum value instead of str.
            call.call_type = CallType.CALL
            return call

    tx = vyper_contract_instance.setNumber(96247783, sender=owner)
    trace = PluginTxTrace(transaction_hash=tx.txn_hash)
    actual = ethereum.enrich_trace(trace)
    assert isinstance(actual, PluginTxTrace)
    assert actual._enriched_calltree is not None
    assert re.match(r"VyperContract\.setNumber\(num=\d*\) \[\d* gas]", repr(actual))

    # Hook into helper method hackily with an enum.
    # Notice the mode is Python and not JSON here.
    call = trace.get_calltree().model_dump(by_alias=True)
    actual = ethereum._enrich_calltree(call)
    assert actual["call_type"] == CallType.CALL.value


def test_enrich_trace_handles_events(ethereum, vyper_contract_instance, owner):
    tx = vyper_contract_instance.setNumber(96247783, sender=owner)

    # Used Hardhat to get the data.
    events = [
        {
            "call_type": "EVENT",
            "data": "0x3ee0dda47a082585c3c66434c4b5b1b4558581efb5f4ee60a59a58bc1201404b00000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000060000000000000000000000000000000000000000000000000000000000000000744796e616d696300000000000000000000000000000000000000000000000000",
            "depth": 1,
            "topics": [
                "0xa84473122c11e32cd505595f246a28418b8ecd6cf819f4e3915363fad1b8f968",
                "0x000000000000000000000000000000000000000000000000000000000000007b",
                "0x9f3d45ac20ccf04b45028b8080bb191eab93e29f7898ed43acf480dd80bba94d",
            ],
        }
    ]

    calldata = "0x3fb5c1cb000000000000000000000000000000000000000000000000000000000000007b"
    call = {
        "events": events,
        "call_type": "CALL",
        "address": vyper_contract_instance.address,
        "calldata": calldata,
    }

    class MyTrace(TransactionTrace):
        def get_calltree(self) -> CallTreeNode:
            return CallTreeNode.model_validate(call)

    trace = MyTrace(transaction_hash=tx.txn_hash)
    actual = ethereum.enrich_trace(trace)
    assert "events" in actual._enriched_calltree, "is evm-trace updated?"
    events = actual._enriched_calltree["events"]
    expected = [
        {
            "name": "NumberChange",
            "calldata": {
                "b": "0x3ee0..404b",
                "prevNum": 0,
                "dynData": '"Dynamic"',
                "newNum": 123,
                "dynIndexed": "0x9f3d..a94d",
            },
        }
    ]
    assert events == expected


def test_enrich_trace_avoids_using_wrong_contract_type(
    ethereum, vyper_contract_instance, solidity_contract_instance, owner, chain
):
    tx = vyper_contract_instance.setNumber(96247783, sender=owner)
    trace = TransactionTrace(transaction_hash=tx.txn_hash)

    # Purposely passing in the wrong contract type here.
    actual = ethereum.enrich_trace(trace, contract_type=solidity_contract_instance.contract_type)

    assert isinstance(actual, TransactionTrace)
    assert actual._enriched_calltree is not None
    assert re.match(r"VyperContract\.setNumber\(num=\d*\) \[\d* gas]", repr(actual))

    # Ensure the contract type was not corrupted.
    ct = chain.contracts[vyper_contract_instance.address]
    assert ct == vyper_contract_instance.contract_type


def test_get_deployment_address(ethereum, owner, vyper_contract_container):
    actual = ethereum.get_deployment_address(owner.address, owner.nonce)
    expected = owner.deploy(vyper_contract_container, 490)
    assert actual == expected



from importlib import import_module

from ape import plugins


@plugins.register(plugins.Config)
def config_class():
    from ape_test.config import ApeTestConfig

    return ApeTestConfig


@plugins.register(plugins.AccountPlugin)
def account_types():
    from ape_test.accounts import TestAccount, TestAccountContainer

    return TestAccountContainer, TestAccount


@plugins.register(plugins.ProviderPlugin)
def providers():
    from ape_test.provider import LocalProvider

    yield "ethereum", "local", LocalProvider


def __getattr__(name: str):
    if name in ("TestAccountContainer", "TestAccount"):
        module = import_module("ape_test.accounts")
    elif name in (
        "EthTesterProviderConfig",
        "GasExclusion",
        "GasConfig",
        "CoverageReportsConfig",
        "CoverageConfig",
        "ApeTestConfig",
    ):
        module = import_module("ape_test.config")
    else:
        module = import_module("ape_test.provider")

    return getattr(module, name)


__all__ = [
    "ApeTestConfig",
    "CoverageConfig",
    "CoverageReportsConfig",
    "EthTesterProviderConfig",
    "GasConfig",
    "GasExclusion",
    "LocalProvider",
    "TestAccount",
    "TestAccountContainer",
]



from tests.integration.cli.utils import run_once


@run_once
def test_cache_init_purge(ape_cli, runner):
    cmd = ("cache", "init", "--network", "ethereum:sepolia")
    result = runner.invoke(ape_cli, cmd, catch_exceptions=False)
    assert "SUCCESS" in result.output
    assert "Caching database initialized for ethereum:sepolia.\n" in result.output

    result = runner.invoke(ape_cli, ("cache", "purge", "--network", "ethereum:sepolia"))
    assert "SUCCESS" in result.output
    assert "Caching database purged for ethereum:sepolia.\n" in result.output



import sys
from abc import abstractmethod
from collections.abc import Iterator, Sequence
from typing import IO, TYPE_CHECKING, Any, Optional

from ape.utils.basemodel import BaseInterfaceModel

if TYPE_CHECKING:
    from ape.types.trace import ContractFunctionPath, GasReport


class TraceAPI(BaseInterfaceModel):
    """
    The class returned from
    :meth:`~ape.api.providers.ProviderAPI.get_transaction_trace`.
    """

    @abstractmethod
    def show(self, verbose: bool = False, file: IO[str] = sys.stdout):
        """
        Show the enriched trace.
        """

    @abstractmethod
    def get_gas_report(
        self, exclude: Optional[Sequence["ContractFunctionPath"]] = None
    ) -> "GasReport":
        """
        Get the gas report.
        """

    @abstractmethod
    def show_gas_report(self, verbose: bool = False, file: IO[str] = sys.stdout):
        """
        Show the gas report.
        """

    @property
    @abstractmethod
    def return_value(self) -> Any:
        """
        The return value deduced from the trace.
        """

    @property
    @abstractmethod
    def revert_message(self) -> Optional[str]:
        """
        The revert message deduced from the trace.
        """

    @abstractmethod
    def get_raw_frames(self) -> Iterator[dict]:
        """
        Get raw trace frames for deeper analysis.
        """

    @abstractmethod
    def get_raw_calltree(self) -> dict:
        """
        Get a raw calltree for deeper analysis.
        """



def test_contract_interaction_in_tests(project, contract_from_fixture, owner):
    contract_in_test = owner.deploy(project.ContractA)
    contract_in_test.setNumber(333, sender=owner)
    actual = contract_in_test.myNumber()
    assert actual == 333

    contract_from_fixture.setNumber(456, sender=owner)
    actual = contract_from_fixture.myNumber()
    assert actual == 456


def test_fail_txn_err_handling(project, contract_from_fixture, owner):
    contract_in_test = owner.deploy(project.ContractA)

    # Errors (uncaught)
    contract_in_test.setNumber(5, sender=owner)



def __getattr__(name: str):
    if name in ("HexBytes",):
        from eth_pydantic_types import HexBytes

        return HexBytes

    elif name in (
        "ABI",
        "Bytecode",
        "Checksum",
        "Compiler",
        "ContractType",
        "PackageManifest",
        "PackageMeta",
        "Source",
    ):
        import ethpm_types

        return getattr(ethpm_types, name)

    elif name in ("Closure",):
        from ethpm_types.source import Closure

        return Closure

    elif name in ("AddressType", "RawAddress"):
        import ape.types.address as address_module

        return getattr(address_module, name)

    elif name in ("HexInt", "_LazySequence"):
        import ape.types.basic as basic_module

        return getattr(basic_module, name)

    elif name in (
        "ContractCoverage",
        "ContractSourceCoverage",
        "CoverageProject",
        "CoverageReport",
        "CoverageStatement",
    ):
        import ape.types.coverage as coverage_module

        return getattr(coverage_module, name)

    elif name in ("ContractLog", "ContractLogContainer", "LogFilter", "MockContractLog"):
        import ape.types.events as events_module

        return getattr(events_module, name)

    elif name in ("AutoGasLimit", "GasLimit"):
        import ape.types.gas as gas_module

        return getattr(gas_module, name)

    elif name in ("MessageSignature", "SignableMessage", "TransactionSignature"):
        import ape.types.signatures as sig_module

        return getattr(sig_module, name)

    elif name in ("ContractFunctionPath", "ControlFlow", "GasReport", "SourceTraceback"):
        import ape.types.trace as trace_module

        return getattr(trace_module, name)

    elif name in ("CurrencyValue", "CurrencyValueComparable"):
        import ape.types.units as units_module

        return getattr(units_module, name)

    elif name in ("BlockID", "ContractCode", "SnapshotID"):
        import ape.types.vm as vm_module

        return getattr(vm_module, name)

    elif name in (
        "BaseInterface",
        "BaseInterfaceModel",
        "BaseModel",
        "ExtraAttributesMixin",
        "ExtraModelAttributes",
        "ManagerAccessMixin",
        "get_attribute_with_extras",
        "get_item_with_extras",
        "only_raise_attribute_error",
    ):
        import ape.utils.basemodel as basemodel_module

        return getattr(basemodel_module, name)

    else:
        raise AttributeError(name)


__all__ = [
    "ABI",
    "AddressType",
    "AutoGasLimit",
    "BaseInterface",
    "BaseInterfaceModel",
    "BaseModel",
    "BlockID",
    "Bytecode",
    "Checksum",
    "Closure",
    "Compiler",
    "ContractCode",
    "ContractCoverage",
    "ContractFunctionPath",
    "ContractLog",
    "ContractLogContainer",
    "ContractSourceCoverage",
    "ContractType",
    "ControlFlow",
    "CoverageProject",
    "CoverageReport",
    "CoverageStatement",
    "CurrencyValue",
    "CurrencyValueComparable",
    "ExtraAttributesMixin",
    "ExtraModelAttributes",
    "GasLimit",
    "GasReport",
    "HexBytes",
    "HexInt",
    "LogFilter",
    "ManagerAccessMixin",
    "MessageSignature",
    "MockContractLog",
    "PackageManifest",
    "PackageMeta",
    "RawAddress",
    "SignableMessage",
    "SnapshotID",
    "Source",
    "SourceTraceback",
    "TransactionSignature",
    "_LazySequence",
    "get_attribute_with_extras",
    "get_item_with_extras",
    "only_raise_attribute_error",
]



from typing import Any

from ape.plugins import Config as RConfig
from ape.plugins import register


@register(RConfig)
def config_class():
    from ape_compile.config import Config

    return Config


def __getattr__(name: str) -> Any:
    if name == "Config":
        from ape_compile.config import Config

        return Config

    else:
        raise AttributeError(name)


__all__ = [
    "Config",
]



import inspect
import os
import sys
import traceback
from contextlib import contextmanager
from pathlib import Path
from runpy import run_module
from typing import Any, Union

import click

from ape.cli.commands import ConnectedProviderCommand
from ape.cli.options import _VERBOSITY_VALUES, _create_verbosity_kwargs, verbosity_option
from ape.exceptions import ApeException, handle_ape_exception
from ape.logging import logger


@contextmanager
def use_scripts_sys_path(path: Path):
    # perf: avoid importing at top of module so `--help` is faster.
    from ape.utils.os import use_temp_sys_path

    # First, ensure there is not an existing scripts module.
    scripts = sys.modules.get("scripts")
    if scripts:
        del sys.modules["scripts"]

        # Sometimes, an installed module is somehow missing a __file__
        # and mistaken by Python to be a system module. We can handle
        # that by also checking some internal properties for the right path.
        excl_paths = (
            [Path(scripts.__file__).parent]
            if scripts.__file__
            else [
                Path(p).parent
                for p in getattr(getattr(scripts, "__path__", None), "_path", []) or []
            ]
        )

    else:
        excl_paths = None

    with use_temp_sys_path(path, exclude=excl_paths):
        yield

    if scripts and sys.modules.get("scripts") != scripts:
        sys.modules["scripts"] = scripts


def get_module(script_path: Path) -> str:
    mod_name = ".".join(
        str(script_path).replace(os.path.sep, ".").split("scripts.")[-1].split(".")[:-1]
    )
    return mod_name if "." in mod_name else f"scripts.{mod_name}"


def run_script_module(script_path: Path):
    mod_name = get_module(script_path)
    return run_module(mod_name)


class ScriptCommand(click.MultiCommand):
    def __init__(self, *args, **kwargs):
        if "result_callback" not in kwargs:
            kwargs["result_callback"] = self.result_callback

        super().__init__(*args, **kwargs)
        self._namespace = {}
        self._command_called = None
        self._has_warned_missing_hook: set[Path] = set()

    def invoke(self, ctx: click.Context) -> Any:
        from ape.utils.basemodel import ManagerAccessMixin as access

        try:
            return super().invoke(ctx)
        except Exception as err:
            if ctx.params["interactive"]:
                # Print the exception trace and then launch the console
                # Attempt to use source-traceback style printing.
                network_value = (
                    ctx.params.get("network") or access.network_manager.default_ecosystem.name
                )
                with access.network_manager.parse_network_choice(
                    network_value, disconnect_on_exit=False
                ):
                    path = ctx.obj.local_project.path
                    assert isinstance(path, Path)  # For mypy.
                    if not isinstance(err, ApeException) or not handle_ape_exception(err, (path,)):
                        err_info = traceback.format_exc()
                        click.echo(err_info)

                    self._launch_console()
            else:
                # Don't handle error - raise exception as normal.
                raise

    def _get_command(self, filepath: Path) -> Union[click.Command, click.Group, None]:
        scripts_folder = Path.cwd() / "scripts"
        relative_filepath = filepath.relative_to(scripts_folder)

        # First load the code module by compiling it
        # NOTE: This does not execute the module
        logger.debug(f"Parsing module: {relative_filepath}")
        try:
            code = compile(filepath.read_text(encoding="utf8"), filepath, "exec")
        except SyntaxError as e:
            logger.error_from_exception(e, f"Exception while parsing script: {relative_filepath}")
            return None  # Prevents stalling scripts

        # NOTE: Introspect code structure only for given patterns (do not execute it to find hooks)
        if "cli" in code.co_names:
            # If the module contains a click cli subcommand, process it and return the subcommand
            logger.debug(f"Found 'cli' command in script: {relative_filepath}")

            with use_scripts_sys_path(filepath.parent.parent):
                try:
                    cli_ns = run_script_module(filepath)
                except Exception as e:
                    logger.error_from_exception(
                        e, f"Exception while parsing script: {relative_filepath}"
                    )
                    return None  # Prevents stalling scripts

            self._namespace[filepath.stem] = cli_ns
            cli_obj = cli_ns["cli"]
            if not isinstance(cli_obj, click.Command):
                logger.warning("Found `cli()` method but it is not a click command.")
                return None

            params = [getattr(x, "name", None) for x in cli_obj.params]
            if "verbosity" not in params:
                option_kwargs = _create_verbosity_kwargs()
                option = click.Option(_VERBOSITY_VALUES, **option_kwargs)
                cli_obj.params.append(option)

            cli_obj.name = filepath.stem if cli_obj.name in ("cli", "", None) else cli_obj.name
            return cli_obj

        elif "main" in code.co_names:
            logger.debug(f"Found 'main' method in script: {relative_filepath}")

            @click.command(
                cls=ConnectedProviderCommand,
                short_help=f"Run '{relative_filepath}:main'",
                name=relative_filepath.stem,
            )
            @verbosity_option()
            def call(network):
                _ = network  # Downstream might use this
                with use_scripts_sys_path(filepath.parent.parent):
                    main_ns = run_script_module(filepath)

                main_ns["main"]()  # Execute the script
                self._namespace[filepath.stem] = main_ns

            return call

        else:
            if relative_filepath not in self._has_warned_missing_hook:
                logger.warning(f"No 'main' method or 'cli' command in script: {relative_filepath}")
                self._has_warned_missing_hook.add(relative_filepath)

            @click.command(
                cls=ConnectedProviderCommand,
                short_help=f"Run '{relative_filepath}:main'",
                name=relative_filepath.stem,
            )
            def call():
                with use_scripts_sys_path(filepath.parent.parent):
                    empty_ns = run_script_module(filepath)

                # Nothing to call, everything executes on loading
                self._namespace[filepath.stem] = empty_ns

            return call

    @property
    def commands(self) -> dict[str, Union[click.Command, click.Group]]:
        # perf: Don't reference `.local_project.scripts_folder` here;
        #   it's too slow when doing just doing `--help`.
        scripts_folder = Path.cwd() / "scripts"
        if not scripts_folder.is_dir():
            return {}

        return self._get_cli_commands(scripts_folder)

    def _get_cli_commands(self, base_path: Path) -> dict:
        commands: dict[str, click.Command] = {}

        for filepath in base_path.iterdir():
            if filepath.stem.startswith("_"):
                continue  # Ignore any "private" files

            elif filepath.is_dir():
                group = click.Group(
                    name=filepath.stem, short_help=f"Run a script from '{filepath.stem}'"
                )
                subcommands = self._get_cli_commands(filepath)
                for subcommand in subcommands.values():
                    group.add_command(subcommand)

                commands[filepath.stem] = group

            if filepath.suffix == ".py":
                cmd = self._get_command(filepath)
                if cmd:  # NOTE: Don't allow calling commands that failed to load
                    commands[filepath.stem] = cmd

        return commands

    def list_commands(self, ctx):
        return sorted(self.commands)

    def get_command(self, ctx, name):
        if Path(name).exists():
            name = Path(name).stem

        if name in self.commands:
            self._command_called = name
            return self.commands[name]

        # NOTE: don't return anything so Click displays proper error

    def result_callback(self, result, interactive: bool):  # type: ignore[override]
        if interactive:
            return self._launch_console()

        return result

    def _launch_console(self):
        from ape.utils.basemodel import ManagerAccessMixin as access

        trace = inspect.trace()
        trace_frames = [
            x for x in trace if x.filename.startswith(str(access.local_project.scripts_folder))
        ]
        if not trace_frames:
            # Error from Ape internals; avoid launching console.
            sys.exit(1)

        # Use most recently matching frame.
        frame = trace_frames[-1].frame

        try:
            globals_dict = {k: v for k, v in frame.f_globals.items() if not k.startswith("__")}
            extra_locals = {
                **self._namespace.get(self._command_called, {}),
                **globals_dict,
                **frame.f_locals,
            }

        finally:
            # Avoid keeping a reference to a frame to avoid reference cycles.
            if frame:
                del frame

        from ape_console._cli import console

        return console(project=access.local_project, extra_locals=extra_locals, embed=True)


@click.command(
    cls=ScriptCommand,
    short_help="Run scripts from the `scripts/` folder",
)
@click.option(
    "-I",
    "--interactive",
    is_flag=True,
    default=False,
    help="Drop into interactive console session after running",
)
def cli(interactive):
    """
    Run scripts from the "scripts/" folder of a project. A script must either define a ``main()``
    method, or define a command named ``cli`` that is a ``click.Command`` or ``click.Group`` object.
    Scripts with only a ``main()`` method will be called with a network option given to the command.
    Scripts with a ``cli`` command should import any mix-ins necessary to operate from the
    ``ape.cli`` package.
    """
    _ = interactive  # NOTE: Used in above callback handler



import re
from abc import ABC
from pathlib import Path
from typing import Optional, cast
from urllib.error import HTTPError, URLError

import pytest
from eth_pydantic_types import HexBytes32
from eth_typing import HexStr
from eth_utils import keccak, to_hex
from hexbytes import HexBytes
from web3 import AutoProvider, Web3
from web3.exceptions import ContractLogicError as Web3ContractLogicError
from web3.exceptions import ExtraDataLengthError
from web3.providers import HTTPProvider

from ape.api import NetworkAPI
from ape.exceptions import (
    APINotImplementedError,
    BlockNotFoundError,
    ConfigError,
    ContractLogicError,
    NetworkMismatchError,
    ProviderError,
    ProviderNotConnectedError,
    TransactionError,
    TransactionNotFoundError,
    VirtualMachineError,
)
from ape.utils import to_int
from ape.utils._web3_compat import ExtraDataToPOAMiddleware
from ape.utils.os import create_tempdir
from ape_ethereum.ecosystem import Block
from ape_ethereum.provider import DEFAULT_SETTINGS, EthereumNodeProvider, Web3Provider
from ape_ethereum.trace import TraceApproach
from ape_ethereum.transactions import (
    AccessList,
    AccessListTransaction,
    DynamicFeeTransaction,
    TransactionStatusEnum,
    TransactionType,
)
from ape_node.provider import GethDev, GethDevProcess, Node, NodeSoftwareNotInstalledError
from tests.conftest import GETH_URI, geth_process_test


@pytest.fixture
def web3_factory(mocker):
    return mocker.patch("ape_ethereum.provider._create_web3")


@pytest.fixture
def process_factory_patch(mocker):
    return mocker.patch("ape_node.provider.GethDevProcess.from_uri")


@pytest.fixture
def tx_for_call(geth_contract):
    return DynamicFeeTransaction.model_validate(
        {
            "chainId": 1337,
            "to": geth_contract.address,
            "gas": 4716984,
            "value": 0,
            "data": HexBytes(keccak(text="myNumber()")[:4]),
            "type": 2,
            "accessList": [],
        }
    )


@geth_process_test
def test_uri(geth_provider):
    assert geth_provider.http_uri == GETH_URI
    assert not geth_provider.ws_uri
    assert geth_provider.uri == GETH_URI


@geth_process_test
def test_uri_when_configured(geth_provider, project, ethereum):
    settings = geth_provider.provider_settings
    geth_provider.provider_settings = {}
    value = "https://value/from/config"
    config = {"node": {"ethereum": {"local": {"uri": value}, "mainnet": {"uri": value}}}}
    expected = DEFAULT_SETTINGS["uri"]
    network = ethereum.get_network("mainnet")

    try:
        with project.temp_config(**config):
            # Assert we use the config value.
            actual_local_uri = geth_provider.uri
            # Assert provider settings takes precedence.
            provider = network.get_provider("node", provider_settings={"uri": expected})
            actual_mainnet_uri = provider.uri

    finally:
        geth_provider.provider_settings = settings

    assert actual_local_uri == value
    assert actual_mainnet_uri == expected


@geth_process_test
def test_uri_non_dev_and_not_configured(mocker, ethereum):
    """
    If the URI was not configured and we are not using a dev
    network (local or -fork), then it should fail, rather than
    use local-host.
    """
    network = ethereum.sepolia.model_copy(deep=True)

    # NOTE: This may fail if using real network names that evmchains would
    #  know about.
    network.name = "gorillanet"
    network.ecosystem.name = "gorillas"

    provider = Node.model_construct(network=network)

    with pytest.raises(ProviderError):
        _ = provider.uri

    # Show that if an evm-chains _does_ exist, it will use that.
    patch = mocker.patch("ape_ethereum.provider.Web3Provider._get_random_rpc")

    # The following URL is made up (please keep example.com).
    expected = "https://gorillas.example.com/v1/rpc"
    patch.return_value = "https://gorillas.example.com/v1/rpc"

    actual = provider.uri
    assert actual == expected


@geth_process_test
def test_uri_invalid(geth_provider, project, ethereum):
    settings = geth_provider.provider_settings
    geth_provider.provider_settings = {}
    value = "I AM NOT A URI OF ANY KIND!"
    config = {"node": {"ethereum": {"local": {"uri": value}}}}

    try:
        with project.temp_config(**config):
            # Assert we use the config value.
            expected = rf"Invalid uri: {re.escape(value)}"
            with pytest.raises(ConfigError, match=expected):
                _ = geth_provider.uri

    finally:
        geth_provider.provider_settings = settings


def test_uri_missing(mock_sepolia):
    class MyProvider(Web3Provider, ABC):
        network: NetworkAPI = mock_sepolia
        name: str = "devnode"
        _connected_uri = None
        _configured_rpc = None

        @property
        def _default_http_uri(self) -> Optional[str]:
            # Doing this to get the error to occur.
            return None

        def connect(self):
            pass

        def disconnect(self):
            pass

    provider = MyProvider()
    with pytest.raises(ProviderError, match="Missing URI for network 'sepolia' on 'ethereum'."):
        _ = provider.uri


@geth_process_test
def test_repr_connected(geth_provider):
    actual = repr(geth_provider)
    expected = f"<Node ({geth_provider.client_version}) chain_id=1337>"
    assert actual == expected


@geth_process_test
def test_repr_on_local_network_and_disconnected(networks):
    node = networks.get_provider_from_choice("ethereum:local:node")
    # Ensure disconnected.
    if w3 := node._web3:
        node._web3 = None

    actual = repr(node)
    expected = "<Node chain_id=1337>"
    assert actual == expected

    if w3:
        node._web3 = w3


@geth_process_test
def test_repr_on_live_network_and_disconnected(networks):
    node = networks.get_provider_from_choice("ethereum:sepolia:node")
    actual = repr(node)
    expected = "<Node chain_id=11155111>"
    assert actual == expected


@geth_process_test
def test_get_logs(geth_contract, geth_account):
    geth_contract.setNumber(101010, sender=geth_account)
    actual = geth_contract.NumberChange[-1]
    assert actual.event_name == "NumberChange"
    assert actual.contract_address == geth_contract.address
    assert actual.event_arguments["newNum"] == 101010


@geth_process_test
def test_chain_id_when_connected(geth_provider):
    assert geth_provider.chain_id == 1337


@geth_process_test
def test_chain_id_live_network_not_connected(networks):
    node = networks.get_provider_from_choice("ethereum:sepolia:node")
    assert node.chain_id == 11155111


@geth_process_test
def test_chain_id_live_network_connected_uses_web3_chain_id(mocker, geth_provider):
    mock_network = mocker.MagicMock()
    mock_network._chain_id = 999999999  # Shouldn't use hardcoded network
    mock_network.name = "mock"
    orig_network = geth_provider.network

    try:
        geth_provider.network = mock_network
        actual = geth_provider.chain_id
    finally:
        geth_provider.network = orig_network

    # Still use the connected chain ID instead network's
    assert actual == 1337


@geth_process_test
def test_chain_id_adhoc_ipc_not_connected(networks, geth_provider):
    ipc_path = str(geth_provider.ipc_path)
    provider = networks.get_provider_from_choice(ipc_path)

    # Clear chain ID for this test.
    provider.__dict__.pop("chain_id", None)
    assert not provider.is_connected, "Provider cannot be connected for this test"

    # We expect this error because this means it attempts RPC to get the chain ID
    # and does not reference the "active" provider (which is different).
    with pytest.raises(ProviderNotConnectedError):
        _ = provider.chain_id


@geth_process_test
def test_connect_wrong_chain_id(ethereum, geth_provider, web3_factory):
    start_network = geth_provider.network
    expected_error_message = (
        f"Provider connected to chain ID '{geth_provider._web3.eth.chain_id}', "
        "which does not match network chain ID '11155111'. "
        "Are you connected to 'sepolia'?"
    )

    try:
        geth_provider.network = ethereum.get_network("sepolia")

        # Ensure when reconnecting, it does not use HTTP
        web3_factory.return_value = geth_provider._web3
        with pytest.raises(NetworkMismatchError, match=expected_error_message):
            geth_provider.connect()
    finally:
        geth_provider.network = start_network


@geth_process_test
def test_connect_to_chain_that_started_poa(mock_web3, web3_factory, ethereum):
    """
    Ensure that when connecting to a chain that
    started out as PoA, such as Sepolia, we include
    the right middleware. Note: even if the chain
    is no longer PoA, we still need the middleware
    to fetch blocks during the PoA portion of the chain.
    """
    mock_web3.eth.get_block.side_effect = ExtraDataLengthError

    def make_request(rpc, arguments):
        if rpc == "eth_chainId":
            return {"result": ethereum.sepolia.chain_id}

        return None

    mock_web3.provider.make_request.side_effect = make_request
    web3_factory.return_value = mock_web3
    provider = ethereum.sepolia.get_provider("node")
    provider.provider_settings = {"uri": "http://node.example.com"}  # fake
    provider.connect()

    # Verify PoA middleware was added.
    assert mock_web3.middleware_onion.inject.call_args[0] == (ExtraDataToPOAMiddleware,)
    assert mock_web3.middleware_onion.inject.call_args[1] == {"layer": 0}


@geth_process_test
def test_connect_using_only_ipc_for_uri_already_connected(project, networks, geth_provider):
    """
    Shows we can remote-connect to a node that is already running when it exposes its IPC path.
    """
    ipc_path = geth_provider.ipc_path
    with project.temp_config(node={"ethereum": {"local": {"uri": f"{ipc_path}"}}}):
        with networks.ethereum.local.use_provider("node") as node:
            assert node.uri == f"{ipc_path}"


@geth_process_test
def test_connect_using_ipc(process_factory_patch, project, networks, geth_provider):
    ipc_path = geth_provider.ipc_path
    with project.temp_config(node={"ethereum": {"local": {"uri": f"{ipc_path}"}}}):
        with networks.ethereum.local.use_provider("node") as node:
            assert node.uri == f"{ipc_path}"


@geth_process_test
def test_connect_request_headers(project, geth_provider, networks):
    http_provider = None
    config = {
        "request_headers": {"h0": 0, "User-Agent": "myapp/2.0"},
        "ethereum": {
            "request_headers": {"h1": 1, "User-Agent": "ETH/1.0"},
            "local": {
                "request_headers": {"h2": 2, "user-agent": "MyPrivateNetwork/0.0.1"},
            },
        },
        "node": {"request_headers": {"h3": 3, "USER-AGENT": "custom-geth-client/v100"}},
    }
    with project.temp_config(**config):
        with networks.ethereum.local.use_provider("node") as geth:
            w3_provider = geth.web3.provider
            if isinstance(w3_provider, AutoProvider):
                for pot_provider_fn in w3_provider._potential_providers:
                    pot_provider = pot_provider_fn()
                    if not isinstance(pot_provider, HTTPProvider):
                        continue
                    else:
                        http_provider = pot_provider

            elif isinstance(w3_provider, HTTPProvider):
                http_provider = w3_provider

            else:
                pytest.fail("Not using HTTP. Please adjust test.")

            assert http_provider is not None, "Setup failed - HTTP Provider still None."

            assert isinstance(http_provider._request_kwargs, dict)
            actual = http_provider._request_kwargs["headers"]
            assert actual["h0"] == 0  # top-level
            assert actual["h1"] == 1  # ecosystem
            assert actual["h2"] == 2  # network
            assert actual["h3"] == 3  # provider

            # Also, assert Ape's default user-agent strings.
            assert actual["User-Agent"].startswith("Ape/")
            assert "Python" in actual["User-Agent"]
            assert "ape-ethereum" in actual["User-Agent"]
            assert "web3.py/" in actual["User-Agent"]

            # Show other default headers.
            assert actual["Content-Type"] == "application/json"

            # Show appended user-agents strings.
            assert "myapp/2.0" in actual["User-Agent"]
            assert "ETH/1.0" in actual["User-Agent"]
            assert "MyPrivateNetwork/0.0.1" in actual["User-Agent"]
            assert "custom-geth-client/v100" in actual["User-Agent"]


@geth_process_test
def test_connect_no_middleware(geth_provider):
    actual = [x for x in geth_provider.web3.middleware_onion]
    # NOTE: There is like 6 default, but for some reason, there is still 1
    #  when not using any, not sure why, but at least we did what we can.
    assert len(actual) <= 1


@geth_process_test
@pytest.mark.parametrize("block_id", (0, "0", "0x0", HexStr("0x0")))
def test_get_block(geth_provider, block_id):
    block = cast(Block, geth_provider.get_block(block_id))

    # Each parameter is the same as requesting the first block.
    assert block.number == 0
    assert block.base_fee == 0
    assert block.gas_used == 0


@geth_process_test
def test_get_block_not_found(geth_provider):
    latest_block = geth_provider.get_block("latest")
    block_id = latest_block.number + 1000
    with pytest.raises(BlockNotFoundError, match=f"Block with ID '{block_id}' not found."):
        geth_provider.get_block(block_id)


@geth_process_test
def test_get_block_pending(geth_provider, geth_account, geth_second_account, accounts):
    """
    Pending timestamps can be weird.
    This ensures we can check those are various strange states of geth.
    """
    actual = geth_provider.get_block("latest")
    assert isinstance(actual, Block)

    snap = geth_provider.snapshot()

    # Transact to increase block
    geth_account.transfer(geth_second_account, "1 gwei")
    actual = geth_provider.get_block("latest")
    assert isinstance(actual, Block)

    # Restore state before transaction
    geth_provider.restore(snap)
    actual = geth_provider.get_block("latest")
    assert isinstance(actual, Block)


@geth_process_test
def test_get_receipt_not_exists_with_timeout(geth_provider):
    txn_hash = "0x0123"
    expected = (
        f"Transaction '{txn_hash}' not found. "
        rf"Error: Transaction '{txn_hash}' "
        "is not in the chain after 0 seconds"
    )
    with pytest.raises(TransactionNotFoundError, match=expected):
        geth_provider.get_receipt(txn_hash, timeout=0)


@geth_process_test
def test_get_receipt(accounts, geth_provider, geth_account, geth_contract):
    receipt = geth_contract.setNumber(111111, sender=geth_account)
    actual = geth_provider.get_receipt(receipt.txn_hash)
    assert receipt.txn_hash == actual.txn_hash
    assert actual.receiver == geth_contract.address
    assert actual.sender == receipt.sender


@geth_process_test
def test_snapshot_and_revert(geth_provider, geth_account, geth_contract):
    snapshot = geth_provider.snapshot()
    start_nonce = geth_account.nonce
    geth_contract.setNumber(211112, sender=geth_account)  # Advance a block
    actual_block_number = geth_provider.get_block("latest").number
    expected_block_number = snapshot + 1
    actual_nonce = geth_account.nonce
    expected_nonce = start_nonce + 1
    assert actual_block_number == expected_block_number
    assert actual_nonce == expected_nonce

    geth_provider.restore(snapshot)

    actual_block_number = geth_provider.get_block("latest").number
    expected_block_number = snapshot
    actual_nonce = geth_account.nonce
    expected_nonce = start_nonce
    assert actual_block_number == expected_block_number
    assert actual_nonce == expected_nonce

    # Use account after revert
    receipt = geth_contract.setNumber(311113, sender=geth_account)  # Advance a block
    assert not receipt.failed


@geth_process_test
def test_return_value_list(geth_account, geth_contract, geth_provider):
    receipt = geth_contract.getFilledArray.transact(sender=geth_account)
    assert receipt.return_value == [1, 2, 3]


@geth_process_test
def test_return_value_nested_address_array(
    geth_account, geth_contract, geth_provider, zero_address
):
    receipt = geth_contract.getNestedAddressArray.transact(sender=geth_account)
    expected = [
        [geth_account.address, geth_account.address, geth_account.address],
        [zero_address, zero_address, zero_address],
    ]
    assert receipt.return_value == expected


@geth_process_test
def test_return_value_nested_struct_in_tuple(geth_account, geth_contract, geth_provider):
    receipt = geth_contract.getNestedStructWithTuple1.transact(sender=geth_account)
    actual = receipt.return_value
    assert actual[0].t.a == geth_account.address
    assert actual[0].foo == 1
    assert actual[1] == 1


@geth_process_test
def test_get_pending_block(geth_provider, geth_account, geth_second_account, accounts):
    """
    Pending timestamps can be weird.
    This ensures we can check those are various strange states of geth.
    """
    actual = geth_provider.get_block("latest")
    assert isinstance(actual, Block)

    snap = geth_provider.snapshot()

    # Transact to increase block
    geth_account.transfer(geth_second_account, "1 gwei")
    actual = geth_provider.get_block("latest")
    assert isinstance(actual, Block)

    # Restore state before transaction
    geth_provider.restore(snap)
    actual = geth_provider.get_block("latest")
    assert isinstance(actual, Block)


@geth_process_test
def test_isolate(chain, geth_contract, geth_account):
    number_at_start = 444
    geth_contract.setNumber(number_at_start, sender=geth_account)
    start_head = chain.blocks.height

    with chain.isolate():
        geth_contract.setNumber(333, sender=geth_account)
        actual = geth_contract.myNumber()
        height = chain.blocks.height

    assert actual == 333
    assert height == start_head + 1
    assert geth_contract.myNumber() == number_at_start

    # Allow extra 1 to account for potential parallelism-related discrepancy
    assert chain.blocks.height in (start_head, start_head + 1)


@geth_process_test
def test_gas_price(geth_provider):
    actual = geth_provider.gas_price
    assert isinstance(actual, int)


@geth_process_test
def test_send_transaction_when_no_error_and_receipt_fails(
    mock_transaction,
    mock_web3,
    geth_provider,
    owner,
    geth_contract,
):
    start_web3 = geth_provider._web3
    geth_provider._web3 = mock_web3
    # Getting a receipt "works", but you get a failed one.
    # NOTE: Value is meaningless.
    tx_hash = HexBytes32.__eth_pydantic_validate__(123**36)
    receipt_data = {
        "failed": True,
        "blockNumber": 0,
        "txnHash": to_hex(tx_hash),
        "status": TransactionStatusEnum.FAILING.value,
        "sender": owner.address,
        "receiver": geth_contract.address,
        "input": b"",
        "gasUsed": 123,
        "gasLimit": 100,
    }

    try:
        # Sending tx "works" meaning no vm error.
        mock_web3.eth.send_raw_transaction.return_value = tx_hash
        mock_web3.eth.wait_for_transaction_receipt.return_value = receipt_data

        # Attempting to replay the tx does not produce any error.
        mock_web3.eth.call.return_value = HexBytes("")

        # Execute test.
        with pytest.raises(TransactionError):
            geth_provider.send_transaction(mock_transaction)

    finally:
        geth_provider._web3 = start_web3


@geth_process_test
def test_send_transaction_exceed_block_gas_limit(chain, geth_provider, geth_contract, geth_account):
    """
    Shows that the local geth node will retry the transaction
    with a new gas if this happens, automatically.
    """
    transaction = geth_contract.setNumber.as_transaction(23333322101, sender=geth_account)
    prepared = geth_account.prepare_transaction(transaction)
    prepared.gas_limit += 100000
    signed = geth_account.sign_transaction(prepared)
    expected_gas_limit = chain.blocks.head.gas_limit
    geth_provider.send_transaction(signed)
    tx_sent = geth_account.history[-1]
    assert tx_sent.gas_limit == expected_gas_limit


@geth_process_test
def test_send_call(geth_provider, ethereum, tx_for_call):
    actual = geth_provider.send_call(tx_for_call)
    assert to_int(actual) == 0


@geth_process_test
def test_send_call_base_class_block_id(networks, ethereum, mocker):
    """
    Testing a case where was a bug in the base class for most providers.
    Note: can't use ape-node as-is, as it overrides `send_call()`.
    """

    provider = mocker.MagicMock()
    provider.network.name = "mainnet"

    def hacked_send_call(*args, **kwargs):
        return EthereumNodeProvider.send_call(provider, *args, **kwargs)

    provider.send_call = hacked_send_call
    tx = ethereum.create_transaction()
    block_id = 567

    orig = networks.active_provider
    networks.active_provider = provider
    _ = provider.send_call(tx, block_id=block_id, skip_trace=True) == HexStr("0x")
    networks.active_provider = orig  # put back ASAP

    actual = provider._prepare_call.call_args[-1]["block_identifier"]
    assert actual == block_id


@geth_process_test
def test_send_call_handles_contract_type_failure(mocker, geth_provider, tx_for_call, mock_web3):
    """
    Fixes an issue where we would get a recursion error during
    handling a CALL failure, which would happen during proxy detection.
    """
    orig_web3 = geth_provider._web3

    def sfx(rpc, arguments, *args, **kwargs):
        if rpc == "eth_call" and arguments[0]:
            raise ContractLogicError()

        return orig_web3.provider.make_request(rpc, arguments, *args, **kwargs)

    # Do this to trigger re-entrancy.
    mock_get = mocker.MagicMock()
    mock_get.side_effect = RecursionError
    orig = geth_provider.chain_manager.contracts.get
    geth_provider.chain_manager.contracts.get = mock_get

    mock_web3.provider.make_request.side_effect = sfx
    geth_provider._web3 = mock_web3
    try:
        with pytest.raises(VirtualMachineError):
            geth_provider.send_call(tx_for_call)
    finally:
        geth_provider._web3 = orig_web3
        geth_provider.chain_manager.contracts.get = orig


@geth_process_test
def test_send_call_skip_trace(mocker, geth_provider, ethereum, tx_for_call):
    """
    When we pass skip_trace=True to `send_call` (as proxy-checking des), we should
    also not bother with any traces in exception handling for that call, as proxy-
    checks fail consistently and getting their traces is unnecessary.
    """
    eth_call_spy = mocker.spy(geth_provider, "_eth_call")
    get_contract_spy = mocker.spy(geth_provider.chain_manager.contracts, "get")
    geth_provider.send_call(tx_for_call, skip_trace=True)
    assert eth_call_spy.call_args[1]["skip_trace"] is True
    assert get_contract_spy.call_count == 0


@geth_process_test
def test_network_choice(geth_provider):
    actual = geth_provider.network_choice
    expected = "ethereum:local:http://127.0.0.1:5550"
    assert actual == expected


@geth_process_test
def test_network_choice_when_custom(geth_provider):
    name = geth_provider.network.name
    geth_provider.network.name = "custom"
    try:
        actual = geth_provider.network_choice
    finally:
        geth_provider.network.name = name

    assert actual == "http://127.0.0.1:5550"


@geth_process_test
def test_make_request_not_exists(geth_provider):
    with pytest.raises(
        APINotImplementedError,
        match="RPC method 'ape_thisDoesNotExist' is not implemented by this node instance.",
    ):
        geth_provider.make_request("ape_thisDoesNotExist")


@geth_process_test
@pytest.mark.parametrize(
    "message",
    (
        "ape_thisDoesNotExist does not exist/is not available",
        "method not found",
        "Method ape_thisDoesNotExist not found",
        "Unknown RPC Endpoint ape_thisDoesNotExist",
    ),
)
def test_make_request_not_exists_different_messages(message, mock_web3, geth_provider):
    def mock_make_request(*args, **kwargs):
        return {"error": message}

    mock_web3.provider.make_request.side_effect = mock_make_request

    class MyProvider(EthereumNodeProvider):
        @property
        def web3(self) -> Web3:
            return mock_web3

    provider = MyProvider(network=geth_provider.network)
    with pytest.raises(
        APINotImplementedError,
        match="RPC method 'ape_thisDoesNotExist' is not implemented by this node instance.",
    ):
        provider.make_request("ape_thisDoesNotExist")


@geth_process_test
def test_geth_bin_not_found():
    bin_name = "__NOT_A_REAL_EXECUTABLE_HOPEFULLY__"
    with pytest.raises(NodeSoftwareNotInstalledError):
        _ = GethDevProcess(Path.cwd() / "notexists", executable=bin_name)


@geth_process_test
def test_base_fee(geth_provider, mocker):
    # NOTE: Only mocked to guarantee we are in a state
    #   with history.
    orig = geth_provider._web3.eth.fee_history
    mock_fee_history = mocker.MagicMock()
    expected = 222
    mock_fee_history.return_value = {"baseFeePerGas": [111, expected]}
    geth_provider._web3.eth.fee_history = mock_fee_history

    try:
        actual = geth_provider.base_fee
    finally:
        geth_provider._web3.eth.fee_history = orig

    assert actual == expected


@geth_process_test
@pytest.mark.parametrize("ret", [{}, {"baseFeePerGas": None}, {"baseFeePerGas": []}])
def test_base_fee_no_history(geth_provider, mocker, ret):
    orig = geth_provider._web3.eth.fee_history
    mock_fee_history = mocker.MagicMock()
    mock_fee_history.return_value = ret
    expected = geth_provider._get_last_base_fee()
    geth_provider._web3.eth.fee_history = mock_fee_history

    try:
        actual = geth_provider.base_fee
    finally:
        geth_provider._web3.eth.fee_history = orig

    assert actual == expected


@geth_process_test
def test_base_fee_hex_decode(geth_provider, mocker):
    orig = geth_provider._web3.eth.fee_history
    mock_fee_history = mocker.MagicMock()
    mock_fee_history.return_value = {"baseFeePerGas": ["0x6", "0x7"]}
    expected = 7
    geth_provider._web3.eth.fee_history = mock_fee_history

    try:
        actual = geth_provider.base_fee
    finally:
        geth_provider._web3.eth.fee_history = orig

    assert actual == expected


@geth_process_test
def test_estimate_gas_cost(geth_contract, geth_provider, geth_account):
    txn = geth_contract.setNumber.as_transaction(900, sender=geth_account)
    estimate = geth_provider.estimate_gas_cost(txn)
    assert estimate > 0


@geth_process_test
def test_estimate_gas_cost_of_static_fee_txn(geth_contract, geth_provider, geth_account):
    txn = geth_contract.setNumber.as_transaction(900, sender=geth_account, type=0)
    estimate = geth_provider.estimate_gas_cost(txn)
    assert estimate > 0


@geth_process_test
def test_estimate_gas_cost_reverts_with_message(geth_contract, geth_provider, geth_second_account):
    # NOTE: The error message from not-owner is "!authorized".
    txn = geth_contract.setNumber.as_transaction(900, sender=geth_second_account, type=0)
    with pytest.raises(ContractLogicError):
        geth_provider.estimate_gas_cost(txn)


@geth_process_test
def test_estimate_gas_cost_reverts_no_message(geth_contract, geth_provider, geth_account):
    # NOTE: The error message from using `5` has no revert message.
    txn = geth_contract.setNumber.as_transaction(5, sender=geth_account, type=0)
    with pytest.raises(ContractLogicError):
        geth_provider.estimate_gas_cost(txn)


@geth_process_test
@pytest.mark.parametrize("tx_type", TransactionType)
def test_prepare_transaction_with_max_gas(tx_type, geth_provider, ethereum, geth_account):
    tx = ethereum.create_transaction(type=tx_type.value, sender=geth_account.address)
    tx.gas_limit = None  # Undo set from validator
    assert tx.gas_limit is None, "Test setup failed - couldn't clear tx gas limit."

    # NOTE: The local network by default uses max_gas.
    actual = geth_provider.prepare_transaction(tx)
    assert actual.gas_limit == geth_provider.max_gas


@geth_process_test
def test_prepare_transaction_access_list_from_rpc(geth_provider, geth_contract, geth_account):
    tx = geth_contract.setNumber.as_transaction(
        123, type=TransactionType.ACCESS_LIST, sender=geth_account.address
    )
    assert isinstance(tx, AccessListTransaction), "Setup failed - should be AccessListTx"
    prepared_tx = geth_provider.prepare_transaction(tx)
    assert isinstance(prepared_tx, AccessListTransaction)
    actual = prepared_tx.access_list
    assert len(actual) > 0
    assert isinstance(actual[0], AccessList)
    assert len(actual[0].storage_keys) > 0


@geth_process_test
def test_create_access_list(geth_provider, geth_contract, geth_account):
    tx = geth_contract.setNumber.as_transaction(123, sender=geth_account, type=1)
    actual = geth_provider.create_access_list(tx)
    assert len(actual) > 0
    assert isinstance(actual[0], AccessList)
    assert len(actual[0].storage_keys) > 0


@geth_process_test
def test_get_virtual_machine_error(geth_provider):
    expected = "__EXPECTED__"
    error = Web3ContractLogicError(f"execution reverted: {expected}", "0x08c379a")
    actual = geth_provider.get_virtual_machine_error(error)
    assert actual.message == expected


@geth_process_test
def test_trace_approach_config(project):
    node_cfg = project.config.node.model_dump(by_alias=True)
    node_cfg["call_trace_approach"] = "geth"
    with project.temp_config(node=node_cfg):
        provider = project.network_manager.ethereum.local.get_provider("node")
        assert provider.call_trace_approach is TraceApproach.GETH_STRUCT_LOG_PARSE


@geth_process_test
def test_start(process_factory_patch, convert, project, geth_provider):
    amount = convert("100_000 ETH", int)

    with project.temp_config(test={"balance": amount}):
        try:
            geth_provider.start()
        except Exception:
            pass  # Exceptions are fine here.

    actual = process_factory_patch.call_args[1]["balance"]
    assert actual == amount


@geth_process_test
@pytest.mark.parametrize("key", ("uri", "ws_uri"))
def test_start_from_ws_uri(process_factory_patch, project, geth_provider, key):
    uri = "ws://localhost:5677"

    settings = geth_provider.provider_settings
    with project.temp_config(node={"ethereum": {"local": {key: uri}}}):
        geth_provider.provider_settings = {}
        try:
            geth_provider.start()
        except Exception:
            pass  # Exceptions are fine here.

        geth_provider.provider_settings = settings

    actual = process_factory_patch.call_args[0][0]  # First "arg"
    assert actual == uri


@geth_process_test
def test_auto_mine(geth_provider, geth_account, geth_contract):
    assert geth_provider.auto_mine is True

    class MyEthereumTestProvider(EthereumNodeProvider):
        """
        Simulates a provider like ape-foundry w/ auto-mine disabled.
        """

        @property
        def auto_mine(self) -> bool:
            return False

    provider = MyEthereumTestProvider(network=geth_provider.network)
    provider._web3 = geth_provider.web3  # Borrow connection.

    tx = geth_contract.setNumber.as_transaction(123)
    tx = geth_account.prepare_transaction(tx)
    tx = geth_account.sign_transaction(tx)
    receipt = provider.send_transaction(tx)
    assert not receipt.confirmed


def test_disconnect_does_not_delete_unrelated_files_in_given_data_dir(networks):
    """
    One time, I used a data-dir containing other files I didn't want to lose. GethDev
    deleted the entire folder during `.disconnect()`, and it was tragic. Ensure this does
    not happen to anyone else.
    """
    with create_tempdir() as temp_dir:
        file = temp_dir / "dont_delete_me_plz.txt"
        file.write_text("Please don't delete me.")

        geth_dev = GethDev(network=networks.ethereum.local)
        geth_dev_proc = GethDevProcess.from_uri(
            "path/to/geth.ipc",
            temp_dir / "geth",
            block_time=1,
            generate_accounts=False,
            initialize_chain=False,
        )
        geth_dev._process = geth_dev_proc
        geth_dev.disconnect()
        assert file.is_file()
        assert not (temp_dir / "geth" / "genesis.json").is_file()


@geth_process_test
def test_ipc_path(geth_provider):
    assert geth_provider.ipc_path.as_posix().endswith("geth.ipc")


class TestGethDevProcess:
    """
    Tests targeting the process-starter directly.
    """

    @pytest.fixture
    def ignore_bin_check(self, mocker):
        # Trick py- into thinking reth is available even when it isn't.
        is_exec_check_patch = mocker.patch("geth.wrapper.is_executable_available")
        is_exec_check_patch.return_value = True

    @geth_process_test
    def test_from_uri_http(self, data_folder):
        geth_dev = GethDevProcess.from_uri("http://localhost:6799", data_folder)
        kwargs = geth_dev.geth_kwargs
        assert kwargs["rpc_addr"] == "localhost"
        assert kwargs["rpc_port"] == "6799"
        assert kwargs["ws_enabled"] is False
        assert kwargs.get("ws_api") is None
        assert kwargs.get("ws_addr") is None
        assert kwargs.get("ws_port") is None

    @geth_process_test
    def test_from_uri_ws(self, data_folder):
        geth_dev = GethDevProcess.from_uri("ws://localhost:6799", data_folder)
        kwargs = geth_dev.geth_kwargs
        assert kwargs.get("rpc_addr") is None
        assert kwargs["ws_enabled"] is True
        assert kwargs["ws_addr"] == "localhost"
        assert kwargs["ws_port"] == "6799"

    @geth_process_test
    def test_from_uri_ipc(self, data_folder):
        geth_dev = GethDevProcess.from_uri("path/to/geth.ipc", data_folder)
        kwargs = geth_dev.geth_kwargs
        assert kwargs["ipc_path"] == "path/to/geth.ipc"
        assert kwargs.get("ws_api") is None
        assert kwargs.get("ws_addr") is None
        assert kwargs.get("rpc_addr") is None

    @geth_process_test
    def test_block_period(self, data_folder):
        geth_dev = GethDevProcess.from_uri(
            "path/to/geth.ipc",
            data_folder,
            block_time=1,
            generate_accounts=False,
            initialize_chain=False,
        )
        assert geth_dev.geth_kwargs["dev_period"] == "1"

    @geth_process_test
    def test_is_rpc_ready_false(self, mocker, data_folder):
        """
        Both Geth and Reth nodes raise simple URLError when the node is not running.
        """
        urlopen_patch = mocker.patch("ape_node.provider.urlopen")
        urlopen_patch.side_effect = URLError("Unable to connect")
        geth_dev = GethDevProcess.from_uri("path/to/geth.ipc", data_folder)
        assert not geth_dev.is_rpc_ready

    @geth_process_test
    def test_is_rpc_ready_true_geth(self, mocker, data_folder):
        """
        Geth has no error when the RPC is ready.
        """
        urlopen_patch = mocker.patch("ape_node.provider.urlopen")
        urlopen_patch.return_value = None
        geth_dev = GethDevProcess.from_uri("path/to/geth.ipc", data_folder)
        assert geth_dev.is_rpc_ready

    @geth_process_test
    def test_is_rpc_ready_true_reth(self, mocker, data_folder):
        """
        Reth raises HTTPError("Method not found") when the RPC is ready.
        """
        urlopen_patch = mocker.patch("ape_node.provider.urlopen")
        urlopen_patch.side_effect = HTTPError("127.0.0.1", 404, "method not found", 0, 0)  # type: ignore
        geth_dev = GethDevProcess.from_uri("path/to/geth.ipc", data_folder)
        assert geth_dev.is_rpc_ready

    @geth_process_test
    def test_command_reth(self, mocker, data_folder, ignore_bin_check):
        """
        Showing we get usable kwargs for a reth --dev node.
        """
        reth_dev = GethDevProcess.from_uri(
            "path/to/reth.ipc", data_folder, executable=["reth", "node"], verify_bin=False
        )
        actual = reth_dev.command
        assert "reth" in actual
        assert "node" in actual
        assert "--http.port" in actual
        assert "--dev" in actual

        # Geth only
        assert "localhost" not in actual
        assert "--maxpeers" not in actual
        assert "--password" not in actual
        assert "--nodiscover" not in actual
        assert "--networkid" not in actual

    @geth_process_test
    def test_ipc_path_geth(self, data_folder):
        geth_dev = GethDevProcess.from_uri("path/to/geth.ipc", data_folder)
        assert geth_dev.ipc_path.endswith("geth.ipc")
        assert geth_dev.geth_kwargs["ipc_path"].endswith("geth.ipc")

    @geth_process_test
    def test_ipc_path_reth(self, data_folder, ignore_bin_check):
        reth_dev = GethDevProcess.from_uri(
            "path/to/reth.ipc", data_folder, executable=["reth", "node"], verify_bin=False
        )
        assert reth_dev.ipc_path.endswith("reth.ipc")
        assert reth_dev.geth_kwargs["ipc_path"].endswith("reth.ipc")

    @geth_process_test
    def test_rpc_api_geth(self, data_folder):
        geth_dev = GethDevProcess.from_uri("path/to/geth.ipc", data_folder)
        actual = set(geth_dev.geth_kwargs["rpc_api"].split(","))
        expected = {"admin", "debug", "eth", "net", "txpool", "web3"}
        assert actual == expected

    @geth_process_test
    def test_rpc_api_reth(self, data_folder, ignore_bin_check):
        reth_dev = GethDevProcess.from_uri(
            "path/to/reth.ipc", data_folder, executable=["reth", "node"], verify_bin=False
        )
        actual = set(reth_dev.geth_kwargs["rpc_api"].split(","))
        expected = {"admin", "debug", "eth", "net", "txpool", "web3", "mev"}
        assert actual == expected



import pytest

from ape.exceptions import APINotImplementedError
from ape.managers.base import BaseManager


@pytest.fixture(scope="module")
def manager():
    class MyManager(BaseManager):
        pass

    return MyManager()


@pytest.mark.parametrize("fn_name", ("_repr_mimebundle_", "_ipython_display_"))
def test_ipython_integration_defaults(manager, fn_name):
    """
    Test default behavior for IPython integration methods.
    The base-manager short-circuits to NotImplementedError to avoid
    dealing with any custom `__getattr__` logic entirely. This prevents
    side-effects such as unnecessary compiling in the ProjectManager.
    """
    with pytest.raises(APINotImplementedError):
        fn = getattr(manager, fn_name)
        fn()



import inspect
from typing import TYPE_CHECKING, Any, Optional

import click

from ape.cli.choices import _NONE_NETWORK, NetworkChoice
from ape.exceptions import NetworkError

if TYPE_CHECKING:
    from click import Context

    from ape.api.networks import ProviderContextManager
    from ape.api.providers import ProviderAPI


def get_param_from_ctx(ctx: "Context", param: str) -> Optional[Any]:
    if value := ctx.params.get(param):
        return value

    # If not found, check the parent context.
    elif parent := ctx.parent:
        return get_param_from_ctx(parent, param)

    return None


def parse_network(ctx: "Context") -> Optional["ProviderContextManager"]:
    from ape.api.providers import ProviderAPI
    from ape.utils.basemodel import ManagerAccessMixin as access

    interactive = get_param_from_ctx(ctx, "interactive")

    # Handle if already parsed (as when using network-option)
    if ctx.obj and "provider" in ctx.obj:
        provider = ctx.obj["provider"]
        return provider.network.use_provider(provider, disconnect_on_exit=not interactive)

    provider = get_param_from_ctx(ctx, "network")
    if provider is not None and isinstance(provider, ProviderAPI):
        return provider.network.use_provider(provider, disconnect_on_exit=not interactive)

    elif provider not in (None, _NONE_NETWORK) and isinstance(provider, str):
        # Is using a choice-str network param value instead of the network object instances.
        return access.network_manager.parse_network_choice(
            provider, disconnect_on_exit=not interactive
        )

    elif provider is None:
        ecosystem = access.network_manager.default_ecosystem
        network = ecosystem.default_network
        if provider_name := network.default_provider_name:
            return network.use_provider(provider_name, disconnect_on_exit=not interactive)
        else:
            raise NetworkError(f"Network {network.name} has no providers.")
    elif provider == _NONE_NETWORK:
        # Was told to skip connection.
        return None
    else:
        raise TypeError(f"Unknown type for network choice: '{provider}'.")


class ConnectedProviderCommand(click.Command):
    """
    A command that uses the :meth:`~ape.cli.options.network_option`.
    It will automatically set the network for the duration of the command execution.
    """

    def __init__(self, *args, **kwargs):
        self._use_cls_types = kwargs.pop("use_cls_types", True)
        self._network_callback = kwargs.pop("network_callback", None)
        super().__init__(*args, **kwargs)

    def parse_args(self, ctx: "Context", args: list[str]) -> list[str]:
        arguments = args  # Renamed for better pdb support.
        base_type: Optional[type] = None if self._use_cls_types else str
        if existing_option := next(
            iter(
                x
                for x in self.params
                if isinstance(x, click.core.Option)
                and x.name == "network"
                and isinstance(x.type, NetworkChoice)
            ),
            None,
        ):
            if base_type is None:
                from ape.api.providers import ProviderAPI

                base_type = ProviderAPI

            # Checking instance above, not sure why mypy still mad.
            existing_option.type.base_type = base_type  # type: ignore

        else:
            # Add the option automatically.
            # NOTE: Local import here only avoids circular import issues.
            from ape.cli.options import NetworkOption

            # NOTE: None base-type will default to `ProviderAPI`.
            option = NetworkOption(base_type=base_type, callback=self._network_callback)
            self.params.append(option)

        return super().parse_args(ctx, arguments)

    def invoke(self, ctx: "Context") -> Any:
        if self.callback is None:
            return

        elif network_ctx := parse_network(ctx):
            with network_ctx as provider:
                return self._invoke(ctx, provider=provider)
        else:
            return self._invoke(ctx)

    def _invoke(self, ctx: "Context", provider: Optional["ProviderAPI"] = None):
        # Will be put back with correct value if needed.
        # Else, causes issues.
        ctx.params.pop("network", None)

        valid_fields = ("ecosystem", "network", "provider")
        requested_fields = (
            []
            if self.callback is None
            else [x for x in inspect.signature(self.callback).parameters if x in valid_fields]
        )
        if self._use_cls_types and requested_fields:
            options = (
                {}
                if provider is None
                else {
                    "ecosystem": provider.network.ecosystem,
                    "network": provider.network,
                    "provider": provider,
                }
            )
            for name in requested_fields:
                if (
                    name not in ctx.params
                    or ctx.params[name] is None
                    or isinstance(ctx.params[name], str)
                ):
                    ctx.params[name] = options.get(name)

        elif not self._use_cls_types and provider is not None:
            # Keep choice-str instead of parsing to objects.
            ctx.params["network"] = provider.network_choice

        return ctx.invoke(self.callback or (lambda: None), **ctx.params)
